/** @file imgfiltg.c
 *  @brief Applies Gaussian filter for dynamic or static PET image data.
 *  @copyright (c) Turku PET Centre
 *  @author Vesa Oikonen, Kaisa Liukko
 */
/// @cond
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
/*****************************************************************************/
#include "libtpcmisc.h"
#include "libtpcimgio.h"
#include "libtpcimgp.h"
#include "libtpcmodext.h"
/*****************************************************************************/

/*****************************************************************************/
static char *info[] = {
  "Applying Gaussian filter for a static or dynamic PET image in",
  "ECAT, NIfTI, or Analyze format. Filtering can be applied to x,y-planes (2D),",
  "x,y,z-data (3D), or across planes (1D).",
  "SD of Gaussian filter must be specified by user, by default in pixels, and",
  "SD=FWHM/2.355.",
  "2D convolution is slow in case of large image and large filter size,",
  "but more accurate than fast 2D and 3D algorithms.",
  " ",
  "Usage: @P [Options] imgfile outputfile SD",
  " ",
  "Options:",
  " -method=<<2D>|<fast2D>|<fast3D>|<fast1D>>",
  "     Specify the Gaussian filtering method; default is the 2D convolution",
  "     method which is slow but accurate; fast2D is a less accurate but fast",
  "     2D Gaussian convolution algorithm with IIR approximation;",
  "     fast3d is 3D Gaussian convolution algorithm with IIR approximation;",
  "     fast1d is 1D Gaussian convolution algorithm with IIR approximation.",
  " -border=<fill|zero>",
  "     Image is extrapolated symmetrically before smoothing (fill, default),",
  "     or data outside image borders is assumed to be zeroes (zero).",
  "     Applies only to 2D method.",
  " -fsize=<filter size>[pxl|mm|cm]",
  "     Size of filter kernel, by default 6*SD; size can be be given in pixels",
  "     (default), mm, or cm. Final size in pixels will be automatically",
  "     changed to the next odd number. Applies only to 2D method.",
  " -steps=<nr>",
  "     More steps provides better accuracy but longer execution time;",
  "     default is 4. Applies only to fast1D, fast2D, and fast3D methods.",
  " -pixelsize=<value>",
  "     Set image pixel size (mm), if missing or wrong in image data.",
  " -stdoptions", // List standard options like --help, -v, etc
  " ",
  "Example #1a: Gaussian SD in pixels",
  "     @P s5998dy1.v s5998dy1_filt.v 6pxl",
  " ",
  "Example #1b: Gaussian with SD of 6 millimetres and 40 mm filter size",
  "     @P -fsize=40mm s5998dy1.v s5998dy1_filt.v 6mm",
  " ",
  "Example #2: Gaussian 3D filtering with SD of 4 pixels",
  "     @P -method=fast3D s5998dy1.v s5998dy1_3dfilt.v 4",
  " ",
  "Example #3: Gaussian 1D filtering across planes, with SD of 3 planes",
  "     @P -method=fast1D s5998dy1.v s5998dy1_zfilt.v 3",
  " ",
  "See also: imgdysmo, imgfsegm, imgthrs, imgbkgrm, fvar4img, imgprofi",
  " ",
  "Keywords: image, smoothing, gaussian, FWHM",
  0};
/*****************************************************************************/

/*****************************************************************************/
/* Turn on the globbing of the command line, since it is disabled by default in
   mingw-w64 (_dowildcard=0); in MinGW32 define _CRT_glob instead, if necessary;
   In Unix&Linux wildcard command line processing is enabled by default. */
/*
#undef _CRT_glob
#define _CRT_glob -1
*/
int _dowildcard = -1;
/*****************************************************************************/

/*****************************************************************************/
/**
 *  Main
 */
int main(int argc, char **argv)
{
  int      ai, help=0, version=0, verbose=1;
  char     petfile[FILENAME_MAX], outfile[FILENAME_MAX];
  int      gaussian_method=1; // 1= 2D convolution; 2= 2D fast IIR; 3= 3D fast IIR; 4= 1D fast IIR
  int      border=1;
  int      kernel_size=0;
  int      filter_unit=0, sd_unit=0; // 0=pixels, 1=mm, 2=cm
  float    filter_sd=-1.0, pixel_size=-1.0, filter_size=-1.0;
  int      step_nr=4;
  int      ret=0;
  char    *cptr;


  /*
   *  Get arguments
   */
  if(argc==1) {tpcPrintUsage(argv[0], info, stderr); return(1);}
  petfile[0]=outfile[0]=(char)0;
  /* Get options */
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') {
    if(tpcProcessStdOptions(argv[ai], &help, &version, &verbose)==0) continue;
    cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(cptr==NULL) continue;
    if(strncasecmp(cptr, "METHOD=", 7)==0 && strlen(cptr)>7) {
      cptr+=7;
      if(strncasecmp(cptr, "2D", 2)==0) {gaussian_method=1; continue;}
      if(strncasecmp(cptr, "FAST1D", 6)==0) {gaussian_method=4; continue;}
      if(strncasecmp(cptr, "FAST2D", 6)==0) {gaussian_method=2; continue;}
      if(strncasecmp(cptr, "FAST3D", 6)==0) {gaussian_method=3; continue;}
    } else if(strncasecmp(cptr, "BORDER=", 7)==0 && strlen(cptr)>7) {
      cptr+=7;
      if(strncasecmp(cptr, "FILL", 1)==0) {border=1; continue;}
      if(strncasecmp(cptr, "ZERO", 1)==0) {border=0; continue;}
      if(strncasecmp(cptr, "YES", 1)==0) {border=1; continue;}
      if(strncasecmp(cptr, "NO", 1)==0) {border=0; continue;}
      if(strcmp(cptr, "1")==0) {border=1; continue;}
      if(strcmp(cptr, "0")==0) {border=0; continue;}
    } else if(strncasecmp(cptr, "FSIZE=", 6)==0) {
      double v;
      cptr+=6; ret=atof_with_check(cptr, &v);
      if(ret==0 && v>0.0) {
        filter_size=(float)v;
        /* Check if unit was specified too */
        if(strcasestr(cptr, "pxl")!=NULL) filter_unit=0;
        if(strcasestr(cptr, "mm")!=NULL) filter_unit=1;
        if(strcasestr(cptr, "cm")!=NULL) filter_unit=2;
        continue;
      }
    } else if(strncasecmp(cptr, "STEPS=", 6)==0) {
      step_nr=atoi(cptr+6); if(step_nr>0) continue;
    } else if(strncasecmp(cptr, "PIXELSIZE=", 10)==0) {
      pixel_size=atof_dpi(cptr+10); if(pixel_size>0) continue;
    }
    fprintf(stderr, "Error: invalid option '%s'.\n", argv[ai]);
    return(1);
  } else break;
  
  /* Print help or version? */
  if(help==2) {tpcHtmlUsage(argv[0], info, ""); return(0);}
  if(help) {tpcPrintUsage(argv[0], info, stdout); return(0);}
  if(version) {tpcPrintBuild(argv[0], stdout); return(0);}
  
  /* Process other arguments, starting from the first non-option */
  if(ai<argc) {strlcpy(petfile, argv[ai], FILENAME_MAX); ai++;}
  if(ai<argc) {strlcpy(outfile, argv[ai], FILENAME_MAX); ai++;}
  if(ai<argc) {
    double v;
    ret=atof_with_check(argv[ai], &v);
    if(ret==0 && v>=0.0) {
      filter_sd=(float)v;
      /* Check if unit was specified too */
      if(strcasestr(argv[ai], "pxl")!=NULL) sd_unit=0;
      if(strcasestr(argv[ai], "mm")!=NULL) sd_unit=1;
      if(strcasestr(argv[ai], "cm")!=NULL) sd_unit=2;
      ai++;
    } else {
      fprintf(stderr, "Error: invalid filter S.D.\n");
      return(1);
    }
  }
  if(ai<argc) {fprintf(stderr, "Error: too many arguments.\n"); return(1);}

  /* Did we get all the information that we need? */
  if(!outfile[0] || filter_sd<0.0) {
    fprintf(stderr, "Error: missing command-line argument; try %s --help\n", argv[0]);
    return(1);
  }
  if(fabs(filter_sd)<1.0E-20) { // allow zero for testing
    filter_sd=0.0;
    fprintf(stderr, "Warning: with given SD no filtering is applied.\n");
  } 
  if(strcasecmp(outfile, petfile)==0) {
    fprintf(stderr, "Error: check the output filenames.\n");
    return(1);
  }


  /* In verbose mode print arguments and options */
  if(verbose>1) {
    printf("petfile := %s\n", petfile);
    printf("outfile := %s\n", outfile);
    if(pixel_size>0.0) printf("pixel_size := %g mm\n", pixel_size);
    printf("gaussian_method := %d\n", gaussian_method);
    printf("filter_sd := %g\n", filter_sd);
    printf("sd_unit := %d\n", sd_unit);
    printf("filter_size := %g\n", filter_size);
    printf("filter_unit := %d\n", filter_unit);
    printf("border := %d\n", border);
    printf("step_nr := %d\n", step_nr);
    fflush(stdout);
  }


  /*
   *  Read the contents of the PET file to img data structure
   */
  IMG img; imgInit(&img);
  if(verbose>0) printf("reading %s\n", petfile);
  ret=imgRead(petfile, &img);
  if(ret) {
    fprintf(stderr, "Error: %s\n", img.statmsg);
    if(verbose>1) printf("ret=%d\n", ret);
    imgEmpty(&img); return(2);
  }
  /* Check if PET data is raw or image */
  if(img.type!=IMG_TYPE_IMAGE) {
    fprintf(stderr, "Error: %s is not an image.\n", petfile);
    imgEmpty(&img); return(2);
  }
  /* Set or check pixel sizes */
  if(pixel_size>0.0) { // given by user
    img.sizex=img.sizey=pixel_size;
  } else {
    /* Check that pixel size is the same in x and y direction */
    pixel_size=img.sizex;
    if(img.sizex!=img.sizey && gaussian_method!=4) {
      fprintf(stderr, "Warning: different pixel x and y size.\n");
      if(img.sizey>pixel_size) pixel_size=img.sizey;
    }
    if(pixel_size<=0.0) {
      if(filter_unit!=0) {
        fprintf(stderr, "Error: unknown pixel size; enter SD in pixels.\n");
        imgEmpty(&img); return(3);
      }
      pixel_size=1.0;
    }
    if(verbose>1) printf("pixel_size := %g\n", pixel_size);
  }
  if(verbose>5) imgInfo(&img);
  /* Check that if 1- or 3-D smoothing is required, image has at least 3 planes */
  if(gaussian_method>2 && img.dimz<3) {
    if(gaussian_method==3) fprintf(stderr, "Error: invalid Z dimension for 3D filtering.\n");
    else fprintf(stderr, "Error: invalid Z dimension for 1D filtering.\n");
    imgEmpty(&img); return(4);
  }


  /*
   *  Set Gaussian SD in units of pixels
   */
  if(sd_unit==2) {filter_sd*=10.0; sd_unit=1;} // cm -> mm
  /* Convert filter SD unit to pixels */
  if(sd_unit!=0) {filter_sd/=pixel_size; sd_unit=0;}
  if(verbose>1) printf("filter_sd_in_pixels := %g\n", filter_sd);


  /*
   *  Gaussian smoothing
   */
  char errmsg[512];
  if(gaussian_method==2) { // fast 2D IIR

    if(verbose>0) fprintf(stdout, "Fast 2D Gaussian filtering\n");
    ret=imgFast2DGaussianFilter(&img, -1, -1, filter_sd, step_nr, verbose-2, errmsg);

  } else if(gaussian_method==3) { // fast 3D IIR

    if(verbose>0) fprintf(stdout, "Fast 3D Gaussian filtering\n");
    ret=imgFast3DGaussianFilter(&img, -1, filter_sd, step_nr, verbose-2, errmsg);

  } else if(gaussian_method==4) { // fast 1D IIR over planes

    if(verbose>0) fprintf(stdout, "Fast 1D Gaussian filtering across planes\n");
    ret=imgFast1DGaussianFilter(&img, filter_sd, step_nr, verbose-2, errmsg);

  } else { // 2D convolution (gaussian_method==1)

    /*
     *  Set the size of filter kernel
     */
    if(filter_size>0.0) { // given by user
      if(filter_unit==2) {filter_size*=10.0; filter_unit=1;} // cm -> mm
      /* Convert filter kernel size unit to pixels */
      if(filter_unit!=0) {filter_size/=pixel_size; filter_unit=0;}
    } else {
      /* Calculate the filter kernel size */
      filter_size=6.0*filter_sd;
      if(verbose>1) printf("filter_size := %g pixels\n", filter_size);
    }
    kernel_size=(int)roundf(filter_size);
    if(kernel_size>=img.dimy) kernel_size=img.dimy;
    if(kernel_size<3) kernel_size=3;
    else if((kernel_size % 2)==0) kernel_size+=1;
    if(verbose>1) printf("kernel_size := %d\n", kernel_size);

    /*
     *  Apply Gaussian filter
     */
    if(verbose>0) fprintf(stdout, "filtering\n");
    ret=imgGaussianFilter(&img, -1, -1, filter_sd, kernel_size, border, verbose-2, errmsg);

  }
  if(ret!=0) {
    fprintf(stderr, "Error in Gaussian filtering: %s.\n", errmsg);
    if(verbose>1) fprintf(stderr, "ret := %d\n", ret);
    imgEmpty(&img); return(8);
  }


  /*
   *  Save filtered image
   */
  if(verbose>1) fprintf(stdout, "writing %s\n", outfile);
  if(imgWrite(outfile, &img)) {
    fprintf(stderr, "Error: %s\n", img.statmsg); 
    imgEmpty(&img);
    return(11);
  }
  if(verbose>0) printf("Written %s\n", outfile);
  imgEmpty(&img);

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/// @endcond
