/** @file simframe.c
 *  @brief Simulates PET time frames with regional TACs.
 *  @remark Previously fr4sim.
 *  @copyright (c) Turku PET Centre
 *  @author Vesa Oikonen
 */
/// @cond
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
#include "libtpcmisc.h"
#include "libtpcmodel.h"
#include "libtpccurveio.h"
#include "libtpcmodext.h"
/*****************************************************************************/

/*****************************************************************************/
static char *info[] = {
  "PET TACs are simulated with short sample time intervals.",
  "This program sums up those points within specified time frames to simulate",
  "a measured PET tissue uptake curve.",
  " ",
  "Usage: @P [options] tacfile framefile newfile [isotope]",
  " ",
  "Options:",
  " -sec or -min",
  "     Sample times are known to be in seconds or minutes, but is not",
  "     specified or is wrong in TAC file.", 
  " -mid",
  "     Frame mid time is written in output file instead of start and end times.",
  " -i  Calculates integrals at frame mid times, instead of averages;",
  "     not available with [isotope].",
  " -ii Calculates 2nd integrals at frame mid times, instead of averages;",
  "     not available with [isotope].",
  " -stdoptions", // List standard options like --help, -v, etc
  " ",
  "TAC file can contain more than one TAC.",
  "Time frames data can be in SIF format. Alternatively, file can consist of",
  "one or two columns of data, containing either 1) frame durations, or",
  "2) frame start times and frame durations; frame time units must be same",
  "as in the datafile. Frames are allowed to overlap.",
  " ",
  "If the isotope is specified, the correction for physical decay is at first",
  "removed, then PET framing is simulated, and after that the framed data",
  "is decay corrected again, based on the frame start time and length as is",
  "the normal procedure when collecting PET image data.",
  "Verify that time units are correct when using this possibility.",
  " ",
  "See also: interpol, tacframe, tac4frpl, tactime, fit2dat, sim_3tcm",
  " ",
  "Keywords: TAC, SIF, simulation, interpolation, time frame",
  0};
/*****************************************************************************/

/*****************************************************************************/
/* Turn on the globbing of the command line, since it is disabled by default in
   mingw-w64 (_dowildcard=0); in MinGW32 define _CRT_glob instead, if necessary;
   In Unix&Linux wildcard command line processing is enabled by default. */
/*
#undef _CRT_glob
#define _CRT_glob -1
*/
int _dowildcard = -1;
/*****************************************************************************/

/*****************************************************************************/
/**
 *  Main
 */
int main(int argc, char **argv)
{
  int        ai, help=0, version=0, verbose=1;
  int        ret, n, frameNr=0;
  int        knownTimeunit=TUNIT_UNKNOWN;
  int        mode=0; // 0=mean, 1=integral, 2=2nd integral
  int        midFrame=0; // save with mid frame times (1) or not (0)
  int        isOverlap=0;
  double    *fstart, *fdur;
  double     lambda=-1.0, hl=-1.0; 
  char       datfile[FILENAME_MAX], frafile[FILENAME_MAX], outfile[FILENAME_MAX];
  char      *cptr, isotopeCode[7];
  DFT        data, data2;
  SIF        sif;


  /*
   *  Get arguments
   */
  if(argc==1) {tpcPrintUsage(argv[0], info, stderr); return(1);}
  dftInit(&data); dftInit(&data2); sifInit(&sif);
  datfile[0]=frafile[0]=outfile[0]=isotopeCode[0]=(char)0;

  /* Options */
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') {
    if(tpcProcessStdOptions(argv[ai], &help, &version, &verbose)==0) continue;
    cptr=argv[ai]+1;
    if(strcasecmp(cptr, "C")==0 || strcasecmp(cptr, "Y")==0) {
      mode=0; continue;
    } else if(strcasecmp(cptr, "I")==0) {
      mode=1; continue;
    } else if(strcasecmp(cptr, "II")==0) {
      mode=2; continue;
    } else if(strncasecmp(cptr, "MINUTES", 3)==0) {
      knownTimeunit=TUNIT_MIN; continue;
    } else if(strncasecmp(cptr, "SECONDS", 3)==0) {
      knownTimeunit=TUNIT_SEC; continue;
    } else if(strcasecmp(cptr, "MID")==0) {
      midFrame=1; continue;
    }
    fprintf(stderr, "Error: invalid option '%s'\n", argv[ai]);
    return(1);
  } else break;

  /* Print help or version? */
  if(help==2) {tpcHtmlUsage(argv[0], info, ""); return(0);}
  if(help) {tpcPrintUsage(argv[0], info, stdout); return(0);}
  if(version) {tpcPrintBuild(argv[0], stdout); return(0);}

  /* Process other arguments, starting from the first non-option */
  if(ai<argc) {strlcpy(datfile, argv[ai], FILENAME_MAX); ai++;}
  if(ai<argc) {strlcpy(frafile, argv[ai], FILENAME_MAX); ai++;}
  if(ai<argc) {strlcpy(outfile, argv[ai], FILENAME_MAX); ai++;}
  if(ai<argc) {
    cptr=hlCorrectIsotopeCode(argv[ai]);
    if(cptr==NULL) {fprintf(stderr, "Error: invalid isotope '%s'\n", argv[ai]); return(1);}
    strcpy(isotopeCode, cptr);
    ai++;
  }
  if(ai<argc) {fprintf(stderr, "Error: invalid argument '%s'\n", argv[ai]); return(1);}

  /* Is something missing? */
  if(!outfile[0]) {tpcPrintUsage(argv[0], info, stdout); return(1);}

  /* Get isotope halflife, if isotope was given by user */
  if(isotopeCode[0]) {
    if(mode>0) {
      fprintf(stderr, "Error: decay option cannot be used with integrals.\n");
      return(1);
    }
    hl=hlFromIsotope(isotopeCode); if(hl<=0.0) {
      fprintf(stderr, "Error: invalid isotope '%s'\n", isotopeCode);
      return(1);
    }
    lambda=hl2lambda(hl);
  }


  /* In verbose mode print arguments and options */
  if(verbose>1) {
    printf("datfile := %s\n", datfile);
    printf("outfile := %s\n", outfile);
    printf("frafile := %s\n", frafile);
    printf("mode := %d\n", mode);
    printf("knownTimeunit := %s\n", dftTimeunit(knownTimeunit));
    printf("midFrame := %d\n", midFrame);
    printf("isotopeCode := %s\n", isotopeCode);
    if(lambda<=0) printf("lambda := %g\n", lambda);
    if(hl>0) printf("halflife := %g\n", hl);
  }

  /*
   *  Read TAC data
   */
  if(verbose>1) printf("reading %s\n", datfile);
  if(dftRead(datfile, &data)) {
    fprintf(stderr, "Error in reading '%s': %s\n", datfile, dfterrmsg);
    dftEmpty(&data);
    return(2);
  }
  /* Correct the timeunit, if given by user */
  if(knownTimeunit!=TUNIT_UNKNOWN) {
    data.timeunit=knownTimeunit;
  } else if(data.timeunit==TUNIT_UNKNOWN) {
    if(data.timeunit==0 && data.x[data.frameNr-1]>30.0 && lambda>0.3) {
      data.timeunit=TUNIT_SEC;
      fprintf(stderr, "Warning: assuming that sample times are in seconds.\n");
    } else {
      data.timeunit=TUNIT_MIN;
      fprintf(stderr, "Warning: assuming that sample times are in minutes.\n");
    }
  }

  /* Check that data and halflife do have the same time unit */
  if(lambda>0) { /* if data in seconds */
    if(data.timeunit==TUNIT_SEC) lambda/=60.;
    if(verbose>1) printf("Lambda := %g\n", lambda);
  }

  /* If necessary, remove decay correction */
  if(lambda>0) {
    int i, j;
    double a=1.0;
    if(verbose>1) printf("removing decay correction.\n");
    for(j=0; j<data.frameNr; j++) {
      a=lambda2factor(-lambda, data.x[j], 0.0);
      for(i=0; i<data.voiNr; i++)
        if(!isnan(data.voi[i].y[j])) data.voi[i].y[j]*=a;
    }
    if(verbose>3) printf("  input data at x=%g multiplied by factor a=%e\n",
      data.x[data.frameNr-1], a);
  }


  /*
   *  Read frame data
   */
  if(verbose>1) printf("reading frame data in %s\n", frafile);
  /* First, try to read it as SIF */
  if(verbose>5) SIF_TEST=1;
  ret=sifRead(frafile, &sif);
  if(ret==0) { // It was SIF
    if(verbose>2) printf("frame times from SIF\n");
    if(verbose>10) sifPrint(&sif);
    frameNr=sif.frameNr;
    fstart = (double*)calloc(frameNr, sizeof(double));
    fdur = (double*)calloc(frameNr, sizeof(double));
    if(fstart==NULL || fdur==NULL) {
      fprintf(stderr, "Error: out of memory.\n");
      sifEmpty(&sif); dftEmpty(&data); return(4);
    }
    /* Copy frame times from SIF */
    for(int i=0; i<frameNr; i++) {
      fstart[i]=sif.x1[i]; fdur[i]=sif.x2[i]-sif.x1[i];
    }
    sifEmpty(&sif);
    /* Convert time units if necessary (always sec in SIF) */
    if(data.timeunit==TUNIT_MIN)
      for(int i=0; i<frameNr; i++) {fstart[i]/=60.0; fdur[i]/=60.0;}
  } else { // It was not SIF, then try to read it otherwise
    if(access(frafile, 0) == -1) {
      fprintf(stderr, "Error: cannot read %s\n", frafile);
      dftEmpty(&data); return(4);
    }
    if(verbose>2) printf("not SIF\n");
    /* Read data */
    IFT ift; iftInit(&ift);
    ret=iftRead(&ift, frafile, 0);
    if(ret) {
      fprintf(stderr, "Error: invalid format in %s\n", frafile);
      dftEmpty(&data); iftEmpty(&ift); return(4);
    }
    /* Go through data lines to compute the number of frame times */
    int vn=0, pn=0;
    double a, b, c;
    n=0;
    for(int ii=0; ii<ift.keyNr; ii++) {
      if(verbose>12) iftWriteItem(&ift, ii, stdout);
      /* do not read comment lines */
      if(ift.item[ii].type!=(char)0 && ift.item[ii].type!=' ') continue;
      /* do not read lines with key */
      if(strlen(ift.item[ii].key)>0) continue;
      /* check that there is a value */
      if(ift.item[ii].value==NULL) continue;
      /* get the nr of values on this line */
      vn=sscanf(ift.item[ii].value, "%lf %lf %lf", &a, &b, &c);
      if(vn<1) continue;
      if(vn>2 || (n>0 && vn!=pn) || (vn==2 && b<=0.0) || (vn==1 && a<=0.0)) {
        fprintf(stderr, "Error: invalid format in %s\n", frafile);
        dftEmpty(&data); iftEmpty(&ift); return(4);
      }
      pn=vn; n++;
    }
    if(n==0) {
      fprintf(stderr, "Error: invalid format in %s\n", frafile);
      dftEmpty(&data); iftEmpty(&ift); return(4);
    }
    /* Allocate memory for frames */
    fstart = (double*)calloc(n, sizeof(double));
    fdur = (double*)calloc(n, sizeof(double));
    if(fstart==NULL || fdur==NULL) {
      fprintf(stderr, "Error: out of memory.\n");
      dftEmpty(&data); iftEmpty(&ift); return(4);
    }
    /* Read frames */
    int i=0;
    for(int ii=0; ii<ift.keyNr && i<n; ii++) {
      if(ift.item[ii].type!=(char)0 && ift.item[ii].type!=' ') continue;
      if(strlen(ift.item[ii].key)>0) continue;
      if(ift.item[ii].value==NULL) continue;
      /* get the nr of values on this line */
      vn=sscanf(ift.item[ii].value, "%lf %lf %lf", &a, &b, &c);
      if(vn<1) continue;
      if(vn==2) {
        fstart[i]=a; fdur[i]=b;
      } else {
        fdur[i]=a;
        if(i==0) fstart[i]=0.0; else fstart[i]=fstart[i-1]+fdur[i-1];
      }
      i++;
    }
    iftEmpty(&ift);
    frameNr=i;
  }
  if(verbose>2) {
    for(int i=0; i<frameNr; i++) {
      printf("frame_start[%d] := %g\n", i+1, fstart[i]);
      printf("frame_duration[%d] := %g\n", i+1, fdur[i]);
    }
  }
  /* Warning, if it looks like time units are different in frame frame */
  if(fstart[frameNr-1]<0.1*data.x[data.frameNr-1] ||
     fstart[frameNr-1]>10.*data.x[data.frameNr-1]) {
    fprintf(stderr, 
            "Warning: PET frame times may not be in same unit as TAC data.\n");
  }

  /* 
   *  Allocate memory for output data, and fill in "header" info 
   */
  ret=dftAllocateWithHeader(&data2, frameNr, data.voiNr, &data);
  if(ret!=0) {
    fprintf(stderr, "Error: cannot setup new TAC data.\n");
    dftEmpty(&data); free(fstart); free(fdur); return(5);
  }
  data2.frameNr=frameNr; data2.voiNr=data.voiNr;

  /* Calculate output frames */
  if(midFrame!=1) data2.timetype=DFT_TIME_STARTEND; 
  else data2.timetype=DFT_TIME_MIDDLE;
  /* Set file format */
  if(data2._type<1) {
    /* ... to PMOD, if extension is .tac, otherwise DFT */
    if(!strcasecmp(filenameGetExtension(outfile), ".tac")) data2._type=DFT_FORMAT_PMOD;
    else data2._type=DFT_FORMAT_STANDARD;
  }
  if(isotopeCode[0]) strcpy(data2.isotope, isotopeCode);
  for(int j=0; j<frameNr; j++) {
    data2.x1[j]=fstart[j]; data2.x2[j]=data2.x1[j]+fdur[j];
    if(j>0 && data2.x2[j-1]>data2.x1[j]) isOverlap=1;
    data2.x[j]=data2.x1[j]+0.5*fdur[j];
    for(int i=0; i<data.voiNr; i++) data2.voi[i].y[j]=0.0;
  }
  free(fstart); free(fdur);
  if(isOverlap) {
    fprintf(stderr, "Warning: time frames overlap.\n");
  }

  /* Check for the need to extrapolate */
  if(lambda>0) { // decay correction goes wrong with extrapolated data
    if(data.x[data.frameNr-1]<0.99*data2.x2[data2.frameNr-1]) {
      fprintf(stderr,
              "Error: cannot extrapolate data when decay is considered.\n");
      dftEmpty(&data); dftEmpty(&data2); return(6);
    } else if(0.95*data.x[0]>data2.x1[0]) {
      double a=0.0;
      /* If mean value of first sample is <=0 then there's no problem */
      for(int i=0; i<data.voiNr; i++) a+=data.voi[i].y[0]; 
      a/=data.voiNr;
      if(a>1.0E-12) {
        fprintf(stderr,
                "Error: cannot extrapolate data when decay is considered.\n");
        if(verbose>1) 
          printf("First data point mean is (%g, %g)\n", data.x[0], a);
        dftEmpty(&data); dftEmpty(&data2); return(6);
      }
    }
  } else { // allow small extrapolation if decay corr is not removed
    if(data.x[data.frameNr-1]<0.9*data2.x2[data2.frameNr-1])
    {
      fprintf(stderr,
              "Error: requested extrapolation too risky.\n");
      dftEmpty(&data); dftEmpty(&data2); return(6);
    }
  }

  /* Interpolate */
  if(verbose>1) printf("interpolating\n");
  for(int i=0, ret=0; i<data2.voiNr; i++) {
    if(mode==0)
      ret=interpolate4pet(data.x, data.voi[i].y, data.frameNr,
        data2.x1, data2.x2, data2.voi[i].y, NULL, NULL, data2.frameNr);
    else if(mode==1)
      ret=interpolate4pet(data.x, data.voi[i].y, data.frameNr,
        data2.x1, data2.x2, NULL, data2.voi[i].y, NULL, data2.frameNr);
    else if(mode==2)
      ret=interpolate4pet(data.x, data.voi[i].y, data.frameNr,
        data2.x1, data2.x2, NULL, NULL, data2.voi[i].y, data2.frameNr);
    if(ret) break;
  }
  if(ret) {
    fprintf(stderr, "Error %d in interpolation.\n", ret);
    dftEmpty(&data); dftEmpty(&data2); return(6);
  }

  /* If necessary, correct again for radioactive decay */
  if(lambda>0) {
    if(verbose>1) printf("adding decay correction.\n");
    double d, a=0.0;
    int i, j;
    for(j=0; j<data2.frameNr; j++) {
      d=data2.x2[j]-data2.x1[j];
      a=lambda2factor(lambda, data2.x1[j], d);
      for(i=0; i<data2.voiNr; i++) data2.voi[i].y[j]*=a;
    }
    if(verbose>3) 
      printf("  last output frame (%g-%g) multiplied by factor a=%e\n",
             data2.x1[data2.frameNr-1], data2.x2[data2.frameNr-1], a);
  }
  dftEmpty(&data);

  /*
   *  Save TACs
   */
  if(verbose>1) printf("saving %s\n", outfile);
  /* Write DFT file */
  ret=dftWrite(&data2, outfile); dftEmpty(&data2);
  if(ret) {
    fprintf(stderr, "Error in writing '%s': %s\n", outfile, dfterrmsg);
    return(11);
  }
  
  /* Tell user what we did */
  if(verbose>=0) {
    if(mode==0) printf("  %d frames saved.\n", frameNr);
    else if(mode==1) printf("  %d integral frames saved.\n", frameNr);
    else if(mode==2) printf("  %d 2nd integral frames saved.\n", frameNr);
  }

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/// @endcond
