/** @file dftdecay.c
 *  @brief Corrects or removes correction for physical decay in PET TAC data.
 *  @copyright (c) Turku PET Centre
 *  @author Vesa Oikonen
 *  @test Test with PMOD format.
 */
/// @cond
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
/*****************************************************************************/
#include "libtpcmisc.h"
#include "libtpccurveio.h"
#include "libtpcmodel.h"
/*****************************************************************************/

/*****************************************************************************/
static char *info[] = {
  "Correction of PET TAC files for physical decay, or removing decay correction.",
  " ",
  "Usage: @P [Options] tacfile [outputfile]",
  " ",
  "Options:",
  " -decay=<on|off>",
  "     Correct the data for radioactive decay (on, default), or remove",
  "     decay correction (off).",
  " -i=<Isotope code>",
  "     If TAC file does not contain the isotope code, it can be specified",
  "     with this option.",
  " -stdoptions", // List standard options like --help, -v, etc
  " ",
  "Example 1: remove the correction for physical decay from a F-18 study,",
  "           writing non-corrected data into a new file.",
  "  @P -decay=off -i=F-18 a123ap.dat a123ap_nodecay.dat",
  " ",
  "Example 2: correct the TAC data for radioactive decay, overwriting the data.",
  "  @P -decay=on -i=C-11 ia456dy1.dft",
  " ",
  "See also: tactime, tacframe, tacunit, tacformat, imgdecay, dcftime",
  " ",
  "Keywords: DFT, input, physical decay, simulation, modelling",
  0};
/*****************************************************************************/

/*****************************************************************************/
/* Turn on the globbing of the command line, since it is disabled by default in
   mingw-w64 (_dowildcard=0); in MinGW32 define _CRT_glob instead, if necessary;
   In Unix&Linux wildcard command line processing is enabled by default. */
/*
#undef _CRT_glob
#define _CRT_glob -1
*/
int _dowildcard = -1;
/*****************************************************************************/

/*****************************************************************************/
/**
 *  Main
 */
int main(int argc, char **argv)
{
  int    ai, help=0, version=0, verbose=1;
  int    decay=1, ret;
  char  *cptr, dftfile1[FILENAME_MAX], dftfile2[FILENAME_MAX], tmp[512];
  char   isotope_name[7];
  DFT    dft;
  double halflife;


  /*
   *  Get arguments
   */
  if(argc==1) {tpcPrintUsage(argv[0], info, stderr); return(1);}
  dftfile1[0]=dftfile2[0]=isotope_name[0]=(char)0;
  halflife=-1.0;
  dftInit(&dft);
  /* Options */
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') {
    cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(cptr==NULL) continue;
    if(tpcProcessStdOptions(argv[ai], &help, &version, &verbose)==0) continue;
    if(strncasecmp(cptr, "I=", 2)==0) {
      strcpy(isotope_name, cptr+2); cptr=hlCorrectIsotopeCode(isotope_name);
      if(cptr!=NULL) {
        strcpy(isotope_name, cptr); halflife=hlFromIsotope(isotope_name);
        if(halflife>0.0) continue;
      }
      fprintf(stderr, "Invalid isotope '%s'\n", isotope_name); return(1);
    } else if(strncasecmp(cptr, "DECAY=", 6)==0) {
      cptr+=6;
      if(strcasecmp(cptr, "ON")==0) {decay=1; continue;}
      else if(strcasecmp(cptr, "OFF")==0) {decay=0; continue;}
    }
    fprintf(stderr, "Error: invalid option '%s'\n", argv[ai]);
    return(1);
  } else break;
  
  /* Print help or version? */
  if(help==2) {tpcHtmlUsage(argv[0], info, ""); return(0);}
  if(help) {tpcPrintUsage(argv[0], info, stdout); return(0);}
  if(version) {tpcPrintBuild(argv[0], stdout); return(0);}

  /* Process other arguments, starting from the first non-option */
  for(; ai<argc; ai++) {
    if(!dftfile1[0]) {
      strcpy(dftfile1, argv[ai]); continue;
    } else if(!dftfile2[0]) {
      strcpy(dftfile2, argv[ai]); continue;
    }
    fprintf(stderr, "Error: invalid argument '%s'\n", argv[ai]);
    return(1);
  }

  /* Is something missing? */
  if(!dftfile1[0]) {
    fprintf(stderr, "Error: missing command-line argument; use option --help\n");
    return(1);
  }
  /* If output filename was not given, use input filename */
  if(!dftfile2[0]) strcpy(dftfile2, dftfile1);


  /* In verbose mode print arguments and options */
  if(verbose>1) {
    printf("dftfile1 := %s\n", dftfile1);
    printf("dftfile2 := %s\n", dftfile2);
    printf("decay := %d\n", decay);
  }


  /*
   *  Read the datafile
   */
  if(verbose>1) printf("reading %s\n", dftfile1);
  if(dftRead(dftfile1, &dft)) {
    fprintf(stderr, "Error in reading '%s': %s\n", dftfile1, dfterrmsg);
    return(2);
  }
  /* If datafile contains valid isotope, then check that it is 
     the same as given with command-line option */
  if(dft.isotope[0]) {
    cptr=hlCorrectIsotopeCode(dft.isotope);
    if(cptr!=NULL && isotope_name[0] && strcasecmp(cptr, isotope_name)!=0 ) {
      fprintf(stderr, "Error: different isotope in %s\n", dftfile1);
      dftEmpty(&dft); return(2);
    }
    if(cptr==NULL && isotope_name[0]) strcpy(dft.isotope, isotope_name);
    if(cptr!=NULL && !isotope_name[0]) {
      strcpy(isotope_name, dft.isotope);
      halflife=hlFromIsotope(dft.isotope);
    }
  } else {
    strcpy(dft.isotope, isotope_name);
  }
  if(verbose>1) {
    printf("isotope := %s\n", isotope_name);
    printf("halflife := %g [min]\n", halflife);
  }
  /* If datafile contains information on decay correction, then check
     that prevent user from decay correcting data twice, or removing
     non-existing decay correction */
  if(dft.decayCorrected!=DFT_DECAY_UNKNOWN) {
    if(decay==1 && dft.decayCorrected==DFT_DECAY_CORRECTED) {
      fprintf(stderr, "Error: physical decay is already corrected in %s.\n", dftfile1);
      dftEmpty(&dft); return(2);
    }
    if(decay==0 && dft.decayCorrected==DFT_DECAY_NOTCORRECTED) {
      fprintf(stderr, "Error: physical decay is not corrected in %s.\n", dftfile1);
      dftEmpty(&dft); return(2);
    }
  }


  /*
   *  If isotope is O-15 and time unit is unknown, check the last sample time
   */
  if(dft.timeunit==TUNIT_UNKNOWN) {
    if(halflife<(HL_O15+0.05)) {
      if(dft.x[dft.frameNr-1]>30.0) {
        dft.timeunit=TUNIT_SEC;
      }
    } else {
      dft.timeunit=TUNIT_MIN;
    }
    fprintf(stderr, "Warning: sample time unit is assumed to be %s\n",
            dftTimeunit(dft.timeunit));
  }
  if(verbose>1) {
    printf("timeunit := %s\n", dftTimeunit(dft.timeunit));
  }

  /*
   *  Decay correction / removal
   */
  ret=dftDecayCorrection(&dft, halflife, decay, 1, 0, 0, tmp, verbose-3);
  if(ret!=0) {
    if(verbose>1) printf("ret := %d\n", ret);
    fprintf(stderr, "Error: %s\n", tmp);
    dftEmpty(&dft); return(5);
  }


  /*
   *  Save corrected file
   */
  dftSetComments(&dft);
  if(dftWrite(&dft, dftfile2)) {
    fprintf(stderr, "Error in writing %s: %s\n", dftfile2, dfterrmsg);
    dftEmpty(&dft); return(11);
  }
  if(verbose>0) {
    if(decay==1) printf("  data was corrected for decay with half-life");
    else printf("  decay correction was removed from data with half-life");
    printf(" %g %s\n", halflife, dftTimeunit(TUNIT_MIN));
    fflush(stdout);
  }

  dftEmpty(&dft);

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/// @endcond
