/** @file csv.c
 *  @brief CSV struct processing.
 *  @test Update test function arguments. 
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpccsv.h"
/*****************************************************************************/

/*****************************************************************************/
/** Initiate the CSV struct before any use.
 *  @author Vesa Oikonen
 *  @sa csvFree, csvAllocate, csvRead
 */
void csvInit(
  /** Pointer to CSV */
  CSV *csv
) {
  if(csv==NULL) return;
  csv->c=NULL; csv->_item_nr=csv->nr=csv->row_nr=csv->col_nr=0;
  csv->separator='\t'; //(char)0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Free memory allocated for CSV data. All contents are destroyed.
 *  @pre Before first use initialize the CSV struct with csvInit().
 *  @author Vesa Oikonen
 *  @sa csvInit, csvAllocate, csvWrite
 */
void csvFree(
  /** Pointer to CSV */
  CSV *csv
) {
  if(csv==NULL) return;
  for(int i=0; i<csv->_item_nr; i++) free(csv->c[i].content);
  free(csv->c);
  csvInit(csv);
}
/*****************************************************************************/

/*****************************************************************************/
/** Allocate memory for list of field items for CSV data.
 *  Any previous content is preserved.
 *  @pre Before first use initialize the CSV struct with csvInit().
 *  @return enum tpcerror (TPCERROR_OK when successful).
 *  @author Vesa Oikonen
 *  @sa csvInit, csvFree
 */
int csvAllocate(
  /** Pointer to CSV */
  CSV *csv,
  /** Nr of field items to add to the list */
  int nr
) {
  if(csv==NULL) return TPCERROR_FAIL;
  if(nr<1) return TPCERROR_OK;

  /* If not allocated previously, then just allocate and thats it */
  if(csv->_item_nr==0) {
    csv->c=(CSV_item*)calloc(nr, sizeof(CSV_item));
    if(csv->c==NULL) return TPCERROR_OUT_OF_MEMORY;
    csv->_item_nr=nr;
    return TPCERROR_OK;
  }

  /* How many unused places there are already? */
  int unused=csv->_item_nr-csv->nr;
  /* Check if enough memory is already allocated */
  if(unused>=nr) return TPCERROR_OK;

  /* Ok, we have to reallocate more memory for the list */
  nr+=csv->nr;
  CSV_item *new_list;
  new_list=(CSV_item*)realloc(csv->c, nr*sizeof(CSV_item));
  if(new_list==NULL) return TPCERROR_OUT_OF_MEMORY;
  else csv->c=new_list;
  /* Initiate the list */
  for(int i=csv->_item_nr; i<nr; i++) {
    csv->c[i].row=0; csv->c[i].col=0; csv->c[i].content=NULL;
  }
  csv->_item_nr=nr;
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Add specified string to CSV.
 *  @return tpcerror (TPCERROR_OK when successful).
 *  @pre Before first use initialize the CSV struct with csvInit().
 *  @author Vesa Oikonen
 *  @sa csvInit, csvPutDouble, csvPutInt, csvPutLine
 */
int csvPutString(
  /** Pointer to initiated CSV; previous contents are not changed */
  CSV *csv,
  /** Field string to add; can be empty ("" or NULL) */
  const char *s,
  /** New line (1) or same line (0) */
  int newline
) {
  if(csv==NULL) return TPCERROR_FAIL;
  /* Allocate memory, if needed, and if needed, then a bigger chunk */
  int avail=csv->_item_nr-csv->nr;
  if(avail<1) {
    int ret=csvAllocate(csv, 10);
    if(ret!=TPCERROR_OK) return ret;    
  }
  /* Write the item */
  if(s==NULL || *s=='\0') csv->c[csv->nr].content=strdup(""); 
  else csv->c[csv->nr].content=strdup(s);
  if(csv->c[csv->nr].content==NULL) return TPCERROR_OUT_OF_MEMORY;
  /* Set the row and column number */
  int r, c;
  if(csv->nr==0) {r=c=0;}
  else {
    if(newline!=0) {
      c=0;
      r=csv->c[csv->nr-1].row+1;
    } else {
      c=csv->c[csv->nr-1].col+1;
      r=csv->c[csv->nr-1].row;
    }
  }
  csv->c[csv->nr].col=c; csv->c[csv->nr].row=r;
  /* Update row_nr and max col nr, if necessary */
  c++; r++;
  if(c>csv->col_nr) csv->col_nr=c;
  if(r>csv->row_nr) csv->row_nr=r;

  csv->nr++;
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Add specified double value as string into CSV.
 *  @return tpcerror (TPCERROR_OK when successful).
 *  @pre Before first use initialize the CSV struct with csvInit().
 *  @author Vesa Oikonen
 *  @sa csvInit, csvPutString, csvPutInt, csvPutLine
 */
int csvPutDouble(
  /** Pointer to initiated CSV; previous contents are not changed */
  CSV *csv,
  /** Double value to add; NaN is added as an empty field */
  double v,
  /** New line (1) or same line (0) */
  int newline,
  /** Convert (1) or do not convert (0) commas to semicolons and
   *  dots to commas */
  int tointl
) {
  int ret;
  if(isnan(v)) {
    ret=csvPutString(csv, "", newline);
  } else {
    char s[128];
    sprintf(s, "%g", v); if(tointl) strReplaceChar(s, '.', ',');
    ret=csvPutString(csv, s, newline);
  }
  return ret;
}
/*****************************************************************************/

/*****************************************************************************/
/** Add specified integer value as string into CSV.
 *  @return tpcerror (TPCERROR_OK when successful).
 *  @pre Before first use initialize the CSV struct with csvInit().
 *  @author Vesa Oikonen
 *  @sa csvInit, csvPutDouble, csvPutString, csvPutLine
 */
int csvPutInt(
  /** Pointer to initiated CSV; previous contents are not changed */
  CSV *csv,
  /** Integer value to add; use csvPutString() to add an empty field */
  int v,
  /** New line (1) or same line (0) */
  int newline
) {
  int ret;
  char s[128];
  sprintf(s, "%d", v);
  ret=csvPutString(csv, s, newline);
  return ret;
}
/*****************************************************************************/

/*****************************************************************************/
/** Count the nr of columns on specified CSV row.
 *  @author Vesa Oikonen
 *  @sa csvCell, csvRead
 *  @return The number of columns.
 */
int csvRowLength(
  /** Pointer to CSV */
  CSV *csv,
  /** CSV row index */
  int row
) {
  if(csv==NULL) return(0);
  int i, n=0;
  for(i=0; i<csv->nr; i++) if(csv->c[i].row==row) n++;
  return(n); 
}
/*****************************************************************************/

/*****************************************************************************/
/** Check whether CSV is regular, that is, each row contain the same number of columns.
 *  @author Vesa Oikonen
 *  @sa csvRowLength, csvRead
 *  @return Returns 1 if CSV is regular, 0 if not.
 */
int csvIsRegular(
  /** Pointer to CSV */
  CSV *csv
) {
  if(csv==NULL) return(0);
  if(csv->nr<2) return(1);
  int i, r, n=0, m=0;
  i=0; r=csv->c[i].row; m++;
  for(i=1; i<csv->nr; i++) {
    if(r==csv->c[i].row) {
      m++; 
      continue;}
    if(n>0 && m!=n) return(0);
    r=csv->c[i].row; n=m; m=1;
  }
  if(n>0 && m!=n) return(0);
  return(1);
}
/*****************************************************************************/

/*****************************************************************************/
/** Get the CVS field contents in specified row and column.
 *  @return Returns pointer to the content string, or NULL if not found.
 *  @author Vesa Oikonen
 *  @sa csvRead, csvSearchField, csvFindField, csvCellReplace
 */
char* csvCell(
  /** Pointer to CSV */
  CSV *csv,
  /** CSV row index */
  int row,
  /** CSV column index */
  int col
) {
  if(csv==NULL) return((char*)NULL);
  for(int i=0; i<csv->nr; i++) 
    if(csv->c[i].row==row && csv->c[i].col==col)
      return(csv->c[i].content);
  return((char*)NULL); 
}
/*****************************************************************************/

/*****************************************************************************/
/** Replace the value of CVS field.
 *  @return tpcerror (TPCERROR_OK when successful).
 *  @author Vesa Oikonen
 *  @sa csvRead, csvSearchField, csvCell
 */
int csvCellReplace(
  /** Pointer to CSV. */
  CSV *csv,
  /** CSV row index. */
  int row,
  /** CSV column index. */
  int col,
  /** Field string to add; can be empty ("" or NULL). */
  const char *s
) {
  if(csv==NULL) return(TPCERROR_FAIL);
  int ci=-1;
  for(int i=0; i<csv->nr; i++) 
    if(csv->c[i].row==row && csv->c[i].col==col) {ci=i; break;}
  if(ci<0) return(TPCERROR_MISSING_DATA);
  /* Delete the previous value */
  free(csv->c[ci].content);
  /* Save the new item */
  if(s==NULL || *s=='\0') csv->c[ci].content=strdup(""); 
  else csv->c[ci].content=strdup(s);
  if(csv->c[ci].content==NULL) return(TPCERROR_OUT_OF_MEMORY);
  return(TPCERROR_OK); 
}
/*****************************************************************************/

/*****************************************************************************/
