/** @file func.c
 *  @brief Mathematical functions evaluation library C file.
 *  @todo Add evaluation functions. 
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpcextensions.h"
/*****************************************************************************/
#include "tpcfunc.h"
/*****************************************************************************/

/*****************************************************************************/
/** Evaluate the function values at specified x values.
 *  @return Returns non-zero in case of an error.
 *  @author Vesa Oikonen
 *  @todo Add support for more functions.
 *  @sa mfEvalInt, mfEvalFrameY
 */
int mfEvalY(
  /** Function code. */
  const char *fid,
  /** Nr of function parameters. */
  const int parNr, 
  /** Array of function parameters. */
  double *p,
  /** Size of x and y arrays. */
  const int sampleNr, 
  /** Array of x values where function values are to be evaluated. */
  double *x,
  /** Pointer to allocated array of y values where function values will be written. */
  double *y,
  /** Verbose level; if zero, then nothing is printed to stderr or stdout. */
  const int verbose
) {
  if(verbose>0) printf("fitEvalY('%s', %d, p[], %d, x[], y[], %d)\n",
                       fid, parNr, sampleNr, verbose);
  if(parNr<1 || p==NULL || sampleNr<1 || x==NULL || y==NULL) return(1);

  int i;
  double xt, delayt=0.0;

  if(strcasecmp(fid, "fengm2")==0) {
    if(parNr<6) return(2);
    if(parNr>6) delayt=p[6];
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; y[i]=0.0; if(!(xt>0.0)) continue;
      y[i] += (p[0]*xt-p[2]-p[4])*exp(p[1]*xt);
      y[i] += p[2]*exp(p[3]*xt);
      y[i] += p[4]*exp(p[5]*xt);
    }
    return(0);
  }
  if(strcasecmp(fid, "fengm2e")==0) {
    if(parNr<8) return(2);
    if(parNr>8) delayt=p[8];
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; y[i]=0.0; if(!(xt>0.0)) continue;
      y[i] += (p[0]*xt-p[2]-p[4]-p[6])*exp(p[1]*xt);
      y[i] += p[2]*exp(p[3]*xt);
      y[i] += p[4]*exp(p[5]*xt);
      y[i] += p[6]*exp(p[7]*xt);
    }
    return(0);
  }
  /* Gamma variate function */
  if(strcasecmp(fid, "gammav")==0) {
    if(parNr<3) return(2);
    if(parNr>3) delayt=p[3];
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; y[i]=0.0; 
      if(!(xt>0.0) || fabs(p[2])<1.0E-10) continue;
      y[i] = p[0]*pow(xt, p[1])*exp(-xt/p[2]);
    }
    return(0);
  }
  /* Exponential function */
  if(strcasecmp(fid, "1exp")==0) {
    if(parNr<2) return(2);
    for(i=0; i<sampleNr; i++) {
      y[i]=0.0; 
      y[i] += p[0]*exp(p[1]*x[i]);
    }
    return(0);
  }
  /* Sum of two exponentials */
  if(strcasecmp(fid, "2exp")==0) {
    if(parNr<4) return(2);
    for(i=0; i<sampleNr; i++) {
      y[i]=0.0; 
      y[i] += p[0]*exp(p[1]*x[i]);
      y[i] += p[2]*exp(p[3]*x[i]);
    }
    return(0);
  }
  /* Sum of three exponentials */
  if(strcasecmp(fid, "3exp")==0) {
    if(parNr<6) return(2);
    for(i=0; i<sampleNr; i++) {
      y[i]=0.0; 
      y[i] += p[0]*exp(p[1]*x[i]);
      y[i] += p[2]*exp(p[3]*x[i]);
      y[i] += p[4]*exp(p[5]*x[i]);
    }
    return(0);
  }
  /* Sum of four exponentials */
  if(strcasecmp(fid, "4exp")==0) {
    if(parNr<8) return(2);
    for(i=0; i<sampleNr; i++) {
      y[i]=0.0; 
      y[i] += p[0]*exp(p[1]*x[i]);
      y[i] += p[2]*exp(p[3]*x[i]);
      y[i] += p[4]*exp(p[5]*x[i]);
      y[i] += p[6]*exp(p[7]*x[i]);
    }
    return(0);
  }
  /* Sum of five exponentials */
  if(strcasecmp(fid, "5exp")==0) {
    if(parNr<10) return(2);
    for(i=0; i<sampleNr; i++) {
      y[i]=0.0; 
      y[i] += p[0]*exp(p[1]*x[i]);
      y[i] += p[2]*exp(p[3]*x[i]);
      y[i] += p[4]*exp(p[5]*x[i]);
      y[i] += p[6]*exp(p[7]*x[i]);
      y[i] += p[8]*exp(p[9]*x[i]);
    }
    return(0);
  }

  /* Weibull cdf with time delay */
  if(strcasecmp(fid, "weibullcdfd")==0) {
    if(parNr<4) return(2);
    delayt=p[0];
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; y[i]=0.0; 
      if(!(xt>0.0)) continue;
      y[i] = p[1]*(1.0-exp(-pow(xt/p[2], p[3])));
    }
    return(0);
  }

  /* Weibull cdf with time delay and cdf derivative */
  if(strcasecmp(fid, "weibullcdfdd")==0) {
    if(parNr<5) return(2);
    delayt=p[0];
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; y[i]=0.0; 
      if(!(xt>0.0)) continue;
      double a=exp(-pow(xt/p[2], p[3]));
      double b=(p[3]/p[2]) * pow(xt/p[2], p[3]-1.0) * a;
      y[i] = p[1]*(b + p[4]*(1.0-a));
    }
    return(0);
  }

  if(verbose>1) printf("function '%s' not supported by fitEvalY()\n", fid);
  return(10);
}
/*****************************************************************************/

/*****************************************************************************/
/** Evaluate the function integral from 0 to specified x values.
 *  @return Returns non-zero in case of an error.
 *  @author Vesa Oikonen
 *  @todo Only MF_FENGM2 and MF_FENGM2E are functional and tested.
 *  @sa mfEvalY, mfEvalFrameY, mfEval2ndInt
 */
int mfEvalInt(
  /** Function code. */
  const char *fid,
  /** Nr of function parameters. */
  const int parNr, 
  /** Array of function parameters. */
  double *p,
  /** Size of x and y arrays. */
  const int sampleNr, 
  /** Array of x values where function values are to be evaluated. */
  double *x,
  /** Pointer to allocated array of integral values where results will be written. */
  double *v,
  /** Verbose level; if zero, then nothing is printed to stderr or stdout. */
  const int verbose
) {
  if(verbose>0) printf("fitEvalInt('%s', %d, p[], %d, x[], v[], %d)\n",
                       fid, parNr, sampleNr, verbose);
  if(parNr<1 || p==NULL || sampleNr<1 || x==NULL || v==NULL) return(1);

  int i;
  double xt, delayt=0.0;

  if(strcasecmp(fid, "fengm2")==0) {
    if(parNr<6) return(2);
    if(parNr>6) delayt=p[6];
    double a;
    double A1, A2, A3, L1, L2, L3;
    A1=p[0]; L1=p[1]; A2=p[2]; L2=p[3]; A3=p[4]; L3=p[5];
    if(fabs(L1)<1.0E-20 || fabs(L2)<1.0E-20 || fabs(L3)<1.0E-20) {
      if(verbose>1) printf("too small L parameter(s)\n");
      return(3);
    }
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; v[i]=0.0; if(!(xt>0.0)) continue;
      if(L1!=0.0) {
        a=exp(L1*xt);
        v[i] += ((A2+A3)/p[1])*(1.0-a);
        v[i] += (A1/(L1*L1))*(1.0+a*(L1*xt - 1.0));
      }
      if(L2!=0.0) v[i]+=(A2/L2)*(exp(L2*xt)-1.0); else v[i]+=A2*xt;
      if(L3!=0.0) v[i]+=(A3/L3)*(exp(L3*xt)-1.0); else v[i]+=A3*xt;
    }
    return(0);
  }
  if(strcasecmp(fid, "fengm2e")==0) {
    if(parNr<8) return(2);
    if(parNr>8) delayt=p[8];
    double a;
    double A1, A2, A3, A4, L1, L2, L3, L4;
    A1=p[0]; L1=p[1]; A2=p[2]; L2=p[3]; A3=p[4]; L3=p[5]; A4=p[6]; L4=p[7];
    if(fabs(L1)<1.0E-20 || fabs(L2)<1.0E-20 || fabs(L3)<1.0E-20 || fabs(L4)<1.0E-20) {
      if(verbose>1) printf("too small L parameter(s)\n");
      return(3);
    }
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; v[i]=0.0; if(!(xt>0.0)) continue;
      if(L1!=0.0) {
        a=exp(L1*xt);
        v[i] += ((A2+A3+A4)/L1)*(1.0-a);
        v[i] += (A1/(L1*L1))*(1.0+a*(L1*xt - 1.0));
      }
      if(L2!=0.0) v[i]+=(A2/L2)*(exp(L2*xt)-1.0); else v[i]+=A2*xt;
      if(L3!=0.0) v[i]+=(A3/L3)*(exp(L3*xt)-1.0); else v[i]+=A3*xt;
      if(L4!=0.0) v[i]+=(A4/L4)*(exp(L4*xt)-1.0); else v[i]+=A4*xt;
    }
    return(0);
  }
  /* Gamma variate function when p[1]==1 */
  if(strcasecmp(fid, "gammav")==0 && parNr>=3 && fabs(p[1]-1.0)<1.0E-10) {
    if(parNr<3) return(2);
    if(parNr>3) delayt=p[3];
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; v[i]=0.0; 
      if(!(xt>0.0) || fabs(p[2])<1.0E-10) continue;
      v[i] = p[0]*p[2]*(p[2] - (p[2]+xt)*exp(-xt/p[2]));
    }
    return(0);
  }

  if(verbose>1) printf("function '%s' not supported by fitEvalInt()\n", fid);
  return(10);
}
/*****************************************************************************/

/*****************************************************************************/
/** Evaluate the 2nd integral of function from 0 to specified x values.
 *  @return Returns non-zero in case of an error.
 *  @todo Only MF_FENGM2 and MF_FENGM2E are functional and tested.
 *  @author Vesa Oikonen
 *  @sa mfEvalInt, mfEvalY, mfEvalFrameY
 */
int mfEval2ndInt(
  /** Function code. */
  const char *fid,
  /** Nr of function parameters. */
  const int parNr, 
  /** Array of function parameters. */
  double *p,
  /** Size of x and y arrays. */
  const int sampleNr, 
  /** Array of x values where function values are to be evaluated. */
  double *x,
  /** Pointer to allocated array of 2nd integral values where results will be written. */
  double *v,
  /** Verbose level; if zero, then nothing is printed to stderr or stdout */
  const int verbose
) {
  if(verbose>0) printf("fitEval2ndInt('%s', %d, p[], %d, x[], v[], %d)\n",
                       fid, parNr, sampleNr, verbose);
  if(parNr<1 || p==NULL || sampleNr<1 || x==NULL || v==NULL) return(1);

  int i;
  double xt, delayt=0.0;

  if(strcasecmp(fid, "fengm2")==0) {
    if(parNr<6) return(2);
    if(parNr>6) delayt=p[6];
    double A1, A2, A3, L1, L2, L3;
    A1=p[0]; L1=p[1]; A2=p[2]; L2=p[3]; A3=p[4]; L3=p[5];
    if(fabs(L1)<1.0E-06 || fabs(L2)<1.0E-06 || fabs(L3)<1.0E-06) {
      if(verbose>1) printf("too small L parameter(s)\n");
      return(3);
    }
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; v[i]=0.0; if(!(xt>0.0)) continue;
      if(L1!=0.0) {
        v[i] += (((A1*xt-A2-A3)*L1 - 2.0*A1) * exp(L1*xt) + 2.0*A1)/(L1*L1*L1);
        v[i] += (A1*xt+A2+A3)/(L1*L1);
        v[i] += (A2*xt+A3*xt)/L1;
      }
      if(L2!=0.0) v[i] -= (A2/(L2*L2))*(1.0 - exp(L2*xt)) + A2*xt/L2;
      if(L3!=0.0) v[i] -= (A3/(L3*L3))*(1.0 - exp(L3*xt)) + A3*xt/L3;
    }
    return(0);
  }
  if(strcasecmp(fid, "fengm2e")==0) {
    if(parNr<8) return(2);
    if(parNr>8) delayt=p[8];
    double A1, A2, A3, A4, L1, L2, L3, L4;
    A1=p[0]; L1=p[1]; A2=p[2]; L2=p[3]; A3=p[4]; L3=p[5]; A4=p[6]; L4=p[7];
    if(fabs(L1)<1.0E-15 || fabs(L2)<1.0E-15 || fabs(L3)<1.0E-15 || fabs(L4)<1.0E-15) {
      if(verbose>1) printf("too small L parameter(s)\n");
      return(3);
    }
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; v[i]=0.0; if(!(xt>0.0)) continue;
      v[i] = A4*L1*L1*L1*L2*L2*L3*L3*L4*xt +
        A4*L1*L1*L1*L2*L2*L3*L3*(1-exp(L4*xt)) +
        L4*L4*(A3*L1*L1*L1*L2*L2*L3*xt + A3*L1*L1*L1*L2*L2*(1.0-exp(L3*xt)) +
          L3*L3*(A2*L1*L1*L1*L2*xt + A2*L1*L1*L1*(1.0-exp(L2*xt)) -
            L2*L2*((A2+A3+A4)*xt*L1*L1 + (A1*xt+A2+A3+A4)*L1 + 
                   ((A1*xt-A2-A3-A4)*L1-2.0*A1)*exp(L1*xt) + 2.0*A1)));
      v[i] /= -(L1*L1*L1*L2*L2*L3*L3*L4*L4);
    }
    return(0);
  }
  /* Gamma variate function when p[1]==1 */
  if(strcasecmp(fid, "gammav")==0 && parNr>=3 && fabs(p[1]-1.0)<1.0E-10) {
    if(parNr<3) return(2);
    if(parNr>3) delayt=p[3];
    for(i=0; i<sampleNr; i++) {
      xt=x[i]-delayt; v[i]=0.0; 
      if(!(xt>0.0) || fabs(p[2])<1.0E-10) continue;
      v[i] = p[0]*p[2]*p[2]*((p[2]+p[2]+xt)*exp(-xt/p[2]) - p[2] - p[2] + xt);
    }
    return(0);
  }

  if(verbose>1) printf("function '%s' not supported by fit2ndEvalInt()\n", fid);
  return(10);
}
/*****************************************************************************/

/*****************************************************************************/
/** Evaluate the PET time frame mean of function values.
 *  @return Returns non-zero in case of an error.
 *  @author Vesa Oikonen
 *  @todo Only MF_FENGM2 and MF_FENGM2E are functional and tested.
 */
int mfEvalFrameY(
  /** Function code. */
  const char *fid,
  /** Nr of function parameters. */
  const int parNr, 
  /** Array of function parameters. */
  double *p,
  /** Size of x and y arrays. */
  const int sampleNr, 
  /** Array of frame start times. */
  double *x1,
  /** Array of frame end times. */
  double *x2,
  /** Pointer to allocated array of y values where function values will be written. */
  double *y,
  /** Verbose level; if zero, then nothing is printed to stderr or stdout. */
  const int verbose
) {
  if(verbose>0) printf("fitEvalFrameY('%s', %d, p[], %d, x1[], x2[], y[], %d)\n",
                       fid, parNr, sampleNr, verbose);
  if(parNr<1 || p==NULL || sampleNr<1 || x1==NULL || x2==NULL || y==NULL) 
    return(1);

  int i, ret;
  double xt1, xt2, v, fd, delayt=0.0;

  if(strcasecmp(fid, "fengm2")==0) {
    if(parNr<6) return(2);
    if(parNr>6) delayt=p[6];
    double A1, A2, A3, L1, L2, L3;
    A1=p[0]; L1=p[1]; A2=p[2]; L2=p[3]; A3=p[4]; L3=p[5];
    if(fabs(L1)<1.0E-12 || fabs(L2)<1.0E-12 || fabs(L3)<1.0E-12) {
      if(verbose>1) printf("too small L parameter(s)\n");
      return(3);
    }
    for(i=0; i<sampleNr; i++) {
      xt1=x1[i]-delayt; xt2=x2[i]-delayt; y[i]=0.0; 
      if(xt1>xt2) {v=xt1; xt1=xt2; xt2=v;}
      if(!(xt2>0.0)) continue;
      fd=xt2-xt1; // requested frame duration
      if(fd<1.0E-025) {
        // if frame duration is about zero, then use the method for that
        ret=mfEvalY(fid, parNr, p, 1, x2+i, y+i, verbose);
        if(ret!=0) return(ret);
      }
      if(!(xt1>0.0)) xt1=0.0; // integral start time must not be <0
      /* Calculate integral between xt1 and xt2 */
      y[i] = A3*L1*L1*L2*(exp(L3*xt1) - exp(L3*xt2)) +
          L3*( A2*L1*L1*(exp(L2*xt1) - exp(L2*xt2)) +
            L2*(((A1*xt1-A2-A3)*L1-A1)*exp(L1*xt1) -  
                ((A1*xt2-A2-A3)*L1-A1)*exp(L1*xt2)
               )
          );
      y[i] /= -(L1*L1*L2*L3);
      /* Divide integral with the original frame duration */
      y[i]/=fd; 
    }
    return(0);
  }
  if(strcasecmp(fid, "fengm2e")==0) {
    if(parNr<8) return(2);
    if(parNr>8) delayt=p[8];
    double A1, A2, A3, A4, L1, L2, L3, L4;
    A1=p[0]; L1=p[1]; A2=p[2]; L2=p[3]; A3=p[4]; L3=p[5]; A4=p[6]; L4=p[7];
    if(fabs(L1)<1.0E-20 || fabs(L2)<1.0E-20 || fabs(L3)<1.0E-20 || fabs(L4)<1.0E-20) {
      if(verbose>1) printf("too small L parameter(s)\n");
      return(3);
    }
    for(i=0; i<sampleNr; i++) {
      xt1=x1[i]-delayt; xt2=x2[i]-delayt; y[i]=0.0; 
      if(xt1>xt2) {v=xt1; xt1=xt2; xt2=v;}
      if(!(xt2>0.0)) continue;
      fd=xt2-xt1; // requested frame duration
      if(fd<1.0E-025) {
        // if frame duration is about zero, then use the method for that
        ret=mfEvalY(fid, parNr, p, 1, x2+i, y+i, verbose);
        if(ret!=0) return(ret);
      }
      if(!(xt1>0.0)) xt1=0.0; // integral start time must not be <0
      /* Calculate integral between xt1 and xt2 */
      y[i] = A4*L1*L1*L2*L3*(exp(L4*xt1) - exp(L4*xt2)) +
        L4*( A3*L1*L1*L2*(exp(L3*xt1) - exp(L3*xt2)) +
          L3*( A2*L1*L1*(exp(L2*xt1) - exp(L2*xt2)) +
            L2*(((A1*xt1-A2-A3-A4)*L1-A1)*exp(L1*xt1) -  
                ((A1*xt2-A2-A3-A4)*L1-A1)*exp(L1*xt2)
               )));
      y[i] /= -(L1*L1*L2*L3*L4);
      /* Divide integral with the original frame duration */
      y[i]/=fd; 
    }
    return(0);
  }
  /* Gamma variate function when p[1]==1 */
  if(strcasecmp(fid, "gammav")==0 && parNr>=3 && fabs(p[1]-1.0)<1.0E-10) {
    if(parNr<3) return(2);
    if(parNr>3) delayt=p[3];
    for(i=0; i<sampleNr; i++) {
      xt1=x1[i]-delayt; xt2=x2[i]-delayt; y[i]=0.0; 
      if(fabs(p[2])<1.0E-10) continue;
      if(xt1>xt2) {v=xt1; xt1=xt2; xt2=v;}
      if(!(xt2>0.0)) continue;
      fd=xt2-xt1; // requested frame duration
      if(fd<1.0E-025) {
        // if frame duration is about zero, then use the method for that
        ret=mfEvalY(fid, parNr, p, 1, x2+i, y+i, verbose);
        if(ret!=0) return(ret);
      }
      if(!(xt1>0.0)) xt1=0.0; // integral start time must not be <0
      /* Calculate integral between xt1 and xt2 */
      y[i] = p[0]*p[2]*( (p[2]+xt1)*exp(-xt1/p[2]) - (p[2]+xt2)*exp(-xt2/p[2]) );
    }
    return(0);
  }

  if(verbose>1) printf("function '%s' not supported by fitEvalFrameY()\n", fid);
  return(10);
}
/*****************************************************************************/

/*****************************************************************************/
