/** @file mpso.c
    @brief Multi particle swarm optimization.
    @copyright (c) Turku PET Centre
    @todo More testing.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpcextensions.h"
#include "tpcrand.h"
#include "tpcstatist.h"
/*****************************************************************************/
#include "tpcnlopt.h"
/*****************************************************************************/

/*****************************************************************************/
/** MPSO particle */
typedef struct MPSO_PARTICLE {
  /** Particle position */
  double *position;
  /** Particle velocity */
  double *velocity;
  /** Cost */
  double cost;
  /** Best position */
  double *bestPosition;
  /** Best cost */
  double bestCost;
} MPSO_PARTICLE;

/** MPSO swarm */
typedef struct MPSO_SWARM {
  /** Particles */
  MPSO_PARTICLE *particle;
  /** Best swarm position */
  double *bestPosition;
  /** Best swarm cost */
  double bestCost;
} MPSO_SWARM;

/** MPSO multiswarm */
typedef struct MPSO_MULTISWARM {
  /** Swarms */
  MPSO_SWARM *swarm;
  /** Best multiswarm position */
  double *bestPosition;
  /** Best multiswarm cost */
  double bestCost;
  /** Dimension */
  unsigned int dim;
} MPSO_MULTISWARM;
/*****************************************************************************/

/*****************************************************************************/
/** Multi particle swarm optimization (MPSO).
    @pre Uses rand(), therefore set seed for a new series of pseudo-random numbers; 
    to produce truly random numbers (not just pseudo-random), do srand(time(0)) before 
    calling this function.
    @post The last objective function call is usually not done with the best parameter estimates;
    if objective function simulates data that you need, you must call the function with the final
    parameters.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
    @todo Better stopping criteria, and options are currently hard-coded.
 */
int nloptMPSO(
  /** Pointer to NLOPT data. 
      Counter funCalls is initially set to zero, and then increased here.
      Parameter maxFunCalls is used as one of the stopping criteria.
      Parameters xtol[] is used as one of the stopping criteria.
   */
  NLOPT *nlo,
  /** Use local optimization (Nelder-Mead) intermittently; using it (1) does not much
      increase the speed than not using it (0), but may increase the success rate. */
  const int doLocal,
  /** Pointer to status data; enter NULL if not needed. */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  if(nlo==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }
  if(nlo->totalNr<1 || nlo->xfull==NULL || nlo->_fun==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }

  if(verbose>1) printf("initializing MPSO\n");
  unsigned int ct=0; // loop index
  unsigned int maxLoops=1500; // max nr of loops (150)
  unsigned int nSwarms=3; // nr of swarms (3)
  unsigned int nParticles=25; // nr of particles (25)
  unsigned int dim; // Nr of parameters
  dim=nlo->totalNr;
  double w=0.333; // inertia (0.333)
  double c1=1.00; // particle / cognitive (1.40)
  double c2=1.60; // swarm / social (1.40)
  double c3=0.40; // multiswarm / global (0.40)
  double death=0.005; ; // probability of particle death-rebirth (0.005)
  double immigrate=0.025;  // probability of particle immigration (0.025)
  /* Initiate the number of function calls */
  nlo->funCalls=0;

  /* Allocate memory */
  MPSO_MULTISWARM multi;
  multi.dim=dim;
  multi.bestPosition=calloc(dim, sizeof(double));
  multi.swarm=calloc(nSwarms, sizeof(MPSO_SWARM));
  for(unsigned int si=0; si<nSwarms; si++) {
    multi.swarm[si].particle=calloc(nParticles, sizeof(MPSO_PARTICLE));
    multi.swarm[si].bestPosition=calloc(dim, sizeof(double));
    for(unsigned int pi=0; pi<nParticles; pi++) {
      multi.swarm[si].particle[pi].position=calloc(dim, sizeof(double));
      multi.swarm[si].particle[pi].velocity=calloc(dim, sizeof(double));
      multi.swarm[si].particle[pi].bestPosition=calloc(dim, sizeof(double));
    }
  }

  /* Check if initial guess is valid */
  int initGuessAvailable=1;
  double initGuessCost=nan("");
  for(unsigned int i=0; i<dim; i++) {
    if(isnan(nlo->xfull[i])) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]<nlo->xlower[i]) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]>nlo->xupper[i]) {initGuessAvailable=0; break;}
  }
  if(initGuessAvailable) {
    initGuessCost=(*nlo->_fun)(dim, nlo->xfull, nlo->fundata);
    nlo->funCalls++;
    if(!isfinite(initGuessCost)) initGuessAvailable=0;
    else if(verbose>1) {
      printf("valid initial guess with cost=%g\n", initGuessCost);
    }
  }

  /* Fill the swarms with random parameters and their costs */
  multi.bestCost=nan("");
  for(unsigned int si=0; si<nSwarms; si++) {
    for(unsigned int pi=0; pi<nParticles; pi++) {
      nloptRandomPoint(multi.swarm[si].particle[pi].position, nlo->xlower, nlo->xupper, dim);
      nloptRandomPoint(multi.swarm[si].particle[pi].velocity, nlo->xlower, nlo->xupper, dim);
      for(unsigned int i=0; i<dim; i++)
        multi.swarm[si].particle[pi].velocity[i]-=multi.swarm[si].particle[pi].position[i];
      multi.swarm[si].particle[pi].cost=
        (*nlo->_fun)(dim, multi.swarm[si].particle[pi].position, nlo->fundata);
      nlo->funCalls++;
      /* Currently, this is the best position of the particle */  
      multi.swarm[si].particle[pi].bestCost=multi.swarm[si].particle[pi].cost;
      doubleCopy(multi.swarm[si].particle[pi].bestPosition, 
                 multi.swarm[si].particle[pi].position, dim);
      /* Set the best cost in this swarm */
      if(pi==0) {
        if(si==0 && initGuessAvailable) { // if initial guess available, use it here
          multi.swarm[si].bestCost=initGuessCost;
          doubleCopy(multi.swarm[si].bestPosition, nlo->xfull, dim);
        } else {
          multi.swarm[si].bestCost=multi.swarm[si].particle[pi].cost;
          doubleCopy(multi.swarm[si].bestPosition, multi.swarm[si].particle[pi].position, dim);
        }
      } else {
        if(multi.swarm[si].bestCost>multi.swarm[si].particle[pi].cost) {
          multi.swarm[si].bestCost=multi.swarm[si].particle[pi].cost;
          doubleCopy(multi.swarm[si].bestPosition, multi.swarm[si].particle[pi].position, dim);
        }
      }
    } // next particle in this swarm
    /* Set the best cost of all swarms */
    if(si==0) {
      multi.bestCost=multi.swarm[si].bestCost;
      doubleCopy(multi.bestPosition, multi.swarm[si].bestPosition, dim);
    } else {
      if(multi.bestCost>multi.swarm[si].bestCost) {
        multi.bestCost=multi.swarm[si].bestCost;
        doubleCopy(multi.bestPosition, multi.swarm[si].bestPosition, dim);
      }
    }
  } // next swarm 

  /* Inside the loop, save the previous best position to check if fit is improving */
  double lastBestPosition[dim];
  unsigned int movementCounter=0; // counter for stopped movement
  unsigned int convergenceCounter=0; // counter for converged swarms


  while(ct<maxLoops && (nlo->maxFunCalls>100 && nlo->funCalls<nlo->maxFunCalls)) {
    ct++; 
    if(verbose>2) printf("-----------------------------\nloop %d\n", ct);
    if(verbose>2) printf("bestCost=%g\n", multi.bestCost);
    if(verbose>3) printf("function_calls_so_far=%d\n", nlo->funCalls);

    /* Go through all swarms and their particles */
    for(unsigned int si=0; si<nSwarms; si++) {

      if(verbose>4) printf("swarm %d bestCost=%g\n", 1+si, multi.swarm[si].bestCost);
      if(verbose>5) {
        printf("  position: %g", multi.swarm[si].bestPosition[0]);
        for(unsigned int i=1; i<dim; i++) printf(", %g", multi.swarm[si].bestPosition[i]);
        printf("\n");
      }

      /* After every 10 loops, try Nelder-Mead, if requested */
      if(doLocal && (ct%10)==0) {
        if(verbose>2) {
          printf("local optimization:\n");
          printf("initial guesses: %g", multi.swarm[si].bestPosition[0]);
          for(unsigned int i=1; i<dim; i++) printf(", %g", multi.swarm[si].bestPosition[i]);
          printf("\nwith cost: %g\n", multi.swarm[si].bestCost);
        }
        doubleCopy(nlo->xfull, multi.swarm[si].bestPosition, dim);
        for(unsigned int i=0; i<dim; i++) {
          double f=(double)(200/ct); if(f<2.0) f=2.0;
          nlo->xdelta[i]=f*nlo->xtol[i];
        }
        if(nloptSimplex(nlo, 50*dim, NULL)==0) {
          double cost=(*nlo->_fun)(dim, nlo->xfull, nlo->fundata);
          if(multi.swarm[si].bestCost>cost) {
            multi.swarm[si].bestCost=cost;
            doubleCopy(multi.swarm[si].bestPosition, nlo->xfull, dim);
            if(verbose>2) {
              printf("resulting parameters: %g", multi.swarm[si].bestPosition[0]);
              for(unsigned int i=1; i<dim; i++) printf(", %g", multi.swarm[si].bestPosition[i]);
              printf("\nwith cost: %g\n", cost);
            }
          } else {
            if(verbose>2) printf("local optimization did not provide lower cost (%g)\n", cost);
          }
        } else {
          if(verbose>2) printf("local optimization failed.\n");
        }
      }

      for(unsigned int pi=0; pi<nParticles; pi++) {
    
        if(verbose>6) printf("particle %d bestCost=%g\n", pi, multi.swarm[si].particle[pi].bestCost);

        /* Cast lots for whether it is time to kill particle and create a new;
           if cost is NaN, then most definitely kill it. */
        if(drand()<death || isnan(multi.swarm[si].particle[pi].cost)) {
          /* Replace current position with random position */
          if(verbose>6) printf("death\n");

          nloptRandomPoint(multi.swarm[si].particle[pi].position, 
                           nlo->xlower, nlo->xupper, dim);
          nloptRandomPoint(multi.swarm[si].particle[pi].velocity, 
                           nlo->xlower, nlo->xupper, dim);
          for(unsigned int i=0; i<dim; i++)
            multi.swarm[si].particle[pi].velocity[i]-=multi.swarm[si].particle[pi].position[i];
          /* cost */
          multi.swarm[si].particle[pi].cost=(*nlo->_fun)(
                     dim, multi.swarm[si].particle[pi].position, nlo->fundata);
          nlo->funCalls++;
          /* Check if this is the best position ever for this particle */  
          if(multi.swarm[si].particle[pi].bestCost>multi.swarm[si].particle[pi].cost) {
            if(verbose>6) printf("Miracle: resurrection led to new best!\n");
            multi.swarm[si].particle[pi].bestCost=multi.swarm[si].particle[pi].cost;
            doubleCopy(multi.swarm[si].particle[pi].bestPosition, 
                       multi.swarm[si].particle[pi].position, dim);
          }
        }

        /* Cast lots for whether it is time to immigrate */
        if(nSwarms>1 && drand()<immigrate) {
          /* Swap particle with a random particle in different swarm */
          if(verbose>6) printf("immigration\n");
          unsigned int sj;
          if(nSwarms<4) {
            sj=si+1; if(sj>=nSwarms) sj=si-1;
          } else {
            do {sj=rand()/(RAND_MAX/(nSwarms-1)+1);} while(si==sj);
          }
          if(verbose>6) printf("  swapping particle between swarms %d and %d\n", 1+si, 1+sj);
          for(unsigned int i=0; i<dim; i++) {
            /* position */
            double d=multi.swarm[si].particle[pi].position[i];
            multi.swarm[si].particle[pi].position[i]=multi.swarm[sj].particle[pi].position[i];
            multi.swarm[sj].particle[pi].position[i]=d;
            /* velocity */
            d=multi.swarm[si].particle[pi].velocity[i];
            multi.swarm[si].particle[pi].velocity[i]=multi.swarm[sj].particle[pi].velocity[i];
            multi.swarm[sj].particle[pi].velocity[i]=d;
          }
          /* cost */
          double d=multi.swarm[si].particle[pi].cost;
          multi.swarm[si].particle[pi].cost=multi.swarm[sj].particle[pi].cost;
          multi.swarm[sj].particle[pi].cost=d;
          /* also the particles memory of its best ever cost and position follows it to another swarm */
          d=multi.swarm[si].particle[pi].bestCost;
          multi.swarm[si].particle[pi].bestCost=multi.swarm[sj].particle[pi].bestCost;
          multi.swarm[sj].particle[pi].bestCost=d;
          for(unsigned int i=0; i<dim; i++) {
            double d=multi.swarm[si].particle[pi].bestPosition[i];
            multi.swarm[si].particle[pi].bestPosition[i]=multi.swarm[sj].particle[pi].bestPosition[i];
            multi.swarm[sj].particle[pi].bestPosition[i]=d;
          }
          /* Is the immigrated particle the best in its new swarm? */
          if(multi.swarm[si].bestCost>multi.swarm[si].particle[pi].bestCost) {
            multi.swarm[si].bestCost=multi.swarm[si].particle[pi].bestCost;
            doubleCopy(multi.swarm[si].bestPosition, multi.swarm[si].particle[pi].bestPosition, dim);
          }
          if(multi.swarm[sj].bestCost>multi.swarm[sj].particle[pi].bestCost) {
            multi.swarm[sj].bestCost=multi.swarm[sj].particle[pi].bestCost;
            doubleCopy(multi.swarm[sj].bestPosition, multi.swarm[sj].particle[pi].bestPosition, dim);
          }
        }

        /* Update velocity in every dimension */
        for(unsigned int i=0; i<dim; i++) {
          multi.swarm[si].particle[pi].velocity[i]=
            w*multi.swarm[si].particle[pi].velocity[i] +
            c1*drand()*(multi.swarm[si].particle[pi].bestPosition[i]-
                        multi.swarm[si].particle[pi].position[i]) +
            c2*drand()*(multi.swarm[si].bestPosition[i]-
                        multi.swarm[si].particle[pi].position[i]) +
            c3*drand()*(multi.bestPosition[i]-
                        multi.swarm[si].particle[pi].position[i]);
        }
        if(verbose>6) {
          printf("  updated velocities: %g", multi.swarm[si].particle[pi].velocity[0]);
          for(unsigned int i=1; i<dim; i++)
            printf(", %g", multi.swarm[si].particle[pi].velocity[i]);
          printf("\n");
        }
      
        /* Update position */
        for(unsigned int i=0; i<dim; i++)
          multi.swarm[si].particle[pi].position[i]+=multi.swarm[si].particle[pi].velocity[i];
        /* check that new position is inside limits */
        for(unsigned int i=0; i<dim; i++)
          if(multi.swarm[si].particle[pi].position[i]<nlo->xlower[i]) {  
            multi.swarm[si].particle[pi].position[i]=nlo->xlower[i];
            /* change direction of velocity */
            multi.swarm[si].particle[pi].velocity[i]*=-0.5; //-1.0;
          } else if(multi.swarm[si].particle[pi].position[i]>nlo->xupper[i]) {  
            multi.swarm[si].particle[pi].position[i]=nlo->xupper[i];
            /* change direction of velocity */
            multi.swarm[si].particle[pi].velocity[i]*=-0.5; //-1.0;
          }
        if(verbose>6) {
          printf("  updated positions: %g", multi.swarm[si].particle[pi].position[0]);
          for(unsigned int i=1; i<dim; i++) printf(", %g", multi.swarm[si].particle[pi].position[i]);
          printf("\n");
        }
        
      
        /* Update cost */
        multi.swarm[si].particle[pi].cost=
          (*nlo->_fun)(dim, multi.swarm[si].particle[pi].position, nlo->fundata);
        nlo->funCalls++;
        if(verbose>6) printf("  updated_cost=%g\n", multi.swarm[si].particle[pi].cost);


        /* Check if this is the best position ever for this particle */  
        if(multi.swarm[si].particle[pi].bestCost>multi.swarm[si].particle[pi].cost) {
          multi.swarm[si].particle[pi].bestCost=multi.swarm[si].particle[pi].cost;
          doubleCopy(multi.swarm[si].particle[pi].bestPosition, 
                     multi.swarm[si].particle[pi].position, dim);
        }

        /* Check if the best position of this particle is the the ever in this swarm?
           Note: check this separately, because of possible immigrations. */
        if(multi.swarm[si].bestCost>multi.swarm[si].particle[pi].bestCost) {
          multi.swarm[si].bestCost=multi.swarm[si].particle[pi].bestCost;
          doubleCopy(multi.swarm[si].bestPosition, multi.swarm[si].particle[pi].bestPosition, dim);
        }

      } // next particle
        
      /* Is the best particle of this swarm the best particle in all swarms? */
      if(multi.bestCost>multi.swarm[si].bestCost) {
        multi.bestCost=multi.swarm[si].bestCost;
        doubleCopy(multi.bestPosition, multi.swarm[si].bestPosition, dim);
      }
      
    } // next swarm

    /* Stop if the best particle in each swarm is inside tolerances of parameters;
       all swarms have then converged to the same position. */
    if(nSwarms>1 && ct>10) {
      int swarmsConverged=1;
      for(unsigned int i=0; i<dim; i++) {
        for(unsigned int si=0; si<nSwarms; si++) {
          double d=fabs(multi.bestPosition[i]-multi.swarm[si].bestPosition[i]);
          if(d>nlo->xtol[i]) {swarmsConverged=0; convergenceCounter=0; break;}
        }
        if(!swarmsConverged) break;
      }
      if(swarmsConverged) {
        convergenceCounter++;
        if(convergenceCounter>5) {
          if(verbose>1) printf("all swarms converged into one minimum.\n");
          break;
        }
      }
    }

    /* Stop if the best overall position stops moving */
    if(ct>=20) {
      int moved=0;
      for(unsigned int i=0; i<dim; i++) {
        double d=fabs(lastBestPosition[i]-multi.bestPosition[i]);
        if(d>nlo->xtol[i]) {moved=1; movementCounter=0; break;}
      }
      if(moved==0) {
        movementCounter++;
        if(movementCounter>14) {
          if(verbose>1) printf("best position did not improve in %d last loops.\n", movementCounter);
          break;
        }
        if(verbose>6 && movementCounter>10) {
          printf("Why does the movement not continue?\n");
          for(unsigned int si=0; si<nSwarms; si++) {
            printf("Swarm %d : bestCost=%g\n", 1+si, multi.swarm[si].bestCost);
            /* Search the currently best particle to get its also its velocity */
            unsigned int pBest=0; 
            double pBestCost=multi.swarm[si].particle[0].cost;
            for(unsigned int pi=1; pi<nParticles; pi++)
              if(multi.swarm[si].particle[pi].cost<pBestCost) {
                pBestCost=multi.swarm[si].particle[pi].cost; pBest=pi;
              }
            printf("  currentBestCost=%g\n", pBestCost);
            printf("\tDim\toPos\tcPos\tcVel\n");
            for(unsigned int i=0; i<dim; i++) {
              printf("\t%d\t%g\t%g\t%g\n", 1+i, multi.swarm[si].bestPosition[i],
                 multi.swarm[si].particle[pBest].position[i], 
                 multi.swarm[si].particle[pBest].velocity[i]);
            }
          }
        }
      }
    }
    doubleCopy(lastBestPosition, multi.bestPosition, dim);

  } // next loop

  /* Check the reason for loop exist */
  if(ct>=maxLoops) {
    if(verbose>1) printf("exceeded the maximum number for loops.\n");
  }
  if(nlo->funCalls>=nlo->maxFunCalls) {
    if(verbose>1) printf("exceeded the maximum number for function calls.\n");
  }

  /* Copy optimized parameters over initial values */
  doubleCopy(nlo->xfull, multi.bestPosition, dim);

  /* Analyze the swarms */
  if(verbose>2) {
    printf("\n---- MPSO end analysis ----\n");
    printf("loops: %d\n", ct);
    printf("function calls: %d\n", nlo->funCalls);
    for(unsigned int si=0; si<nSwarms; si++) {
      printf("Swarm %d : bestCost=%g\n", 1+si, multi.swarm[si].bestCost);
      for(unsigned int i=0; i<dim; i++) {
        double plist[nParticles], vlist[nParticles], pmean, vmean, psd, vsd;
        for(unsigned int pi=0; pi<nParticles; pi++) {
          plist[pi]=multi.swarm[si].particle[pi].position[i];
          vlist[pi]=multi.swarm[si].particle[pi].velocity[i];
        }
        statMeanSD(plist, nParticles, &pmean, &psd, NULL);
        statMeanSD(vlist, nParticles, &vmean, &vsd, NULL);
        printf("parameter %d: position %g (%g+-%g)  velocity %g+-%g\n", 
               1+i, multi.swarm[si].bestPosition[i], pmean, psd, vmean, vsd);
      }
    }
    printf("Overall bestCost=%g\n", multi.bestCost);
    printf("at position: %g", multi.bestPosition[0]);
    for(unsigned int i=1; i<dim; i++) printf(", %g", multi.bestPosition[i]);
    printf("\n");
  }
  
  /* Free allocated memory */
  for(unsigned int si=0; si<nSwarms; si++) {
    for(unsigned int pi=0; pi<nParticles; pi++) {
      free(multi.swarm[si].particle[pi].position);
      free(multi.swarm[si].particle[pi].velocity);
      free(multi.swarm[si].particle[pi].bestPosition);
    }
    free(multi.swarm[si].particle);
    free(multi.swarm[si].bestPosition);
  }
  free(multi.bestPosition);
  free(multi.swarm);


  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
