/// @file tgo.c
/// @brief Topographical global optimization algorithm.
/// @author Vesa Oikonen
/// @copyright (c) Turku PET Centre
///
/// Based on an algorithm by Aimo Törn and Sami Viitanen. References: 
/// - Törn A, Viitanen S. Iterative topographical global optimization. In:
///   C.A. Floudas and P.M. Pardalos (eds.) State of the Art in Global Optimization.
///   Kluwer Academic Publishers, 1996, pp 353-363.
/// - http://www.abo.fi/~atorn/ProbAlg/Page53.html
///
/// @remark Not well tested!
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpcextensions.h"
#include "tpcrand.h"
#include "tpcstatist.h"
/*****************************************************************************/
#include "tpcnlopt.h"
/*****************************************************************************/

/*****************************************************************************/
#ifndef MAX_PARAMETERS
/** Max nr of parameters */
#define MAX_PARAMETERS 50
#endif
/** One TGO point */
typedef struct {
  /** Graphic minimum (1) or not (0). */
  int topomin;
  /** Value of objective function. */
  double fvalue;
  /** Parameters for objective function. */
  double par[MAX_PARAMETERS];
} TGO_POINT;
/*****************************************************************************/

/*****************************************************************************/
/** Iterative Topographical global optimization algorithm 1 (iTGO1).

    @remark Not well tested!
    @pre Uses rand(), therefore set seed for a new series of pseudo-random numbers; 
    to produce truly random numbers (not just pseudo-random), do srand(time(0)) before 
    calling this function.
    @post The last objective function call is usually not done with the best parameter estimates;
    if objective function simulates data that you need, you must call the function with the final
    parameters.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
    @todo Finalize the code and tests.
 */
int nloptITGO1(
  /** Pointer to NLOPT data. 
      Counter funCalls is initially set to zero, and then increased here.
      Parameter maxFunCalls is used as one of the stopping criteria.
      Parameters xtol[] is used for scaling and as one of the stopping criteria.
   */
  NLOPT *nlo,
  /** Local optimization method: 0=no local optimization, 1=Nelder-Mead. */
  const int doLocal,
  /** Number of TGO iterations; enter 0 to use the default; enter 1 to run TGO
      just once, ie to use TGO instead of iTGO. Iterations are needed if sampleNr
      (below) would otherwise require too much memory. */
  unsigned int tgoNr,
  /** Number of points to sample in one iteration; may need to be large if the number
      of iterations (above) is small; enter 0 to let function decide it. */
  unsigned int sampleNr,
  /** Number of neighbours to investigate (cluster size); enter a large number relative 
      to sampleNr (above) if only few local minima are expected; enter 0 to let
      function decide it. */
  unsigned int neighNr,
  /** Pointer to status data; enter NULL if not needed. */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) {
    printf("%s(NLOPT, %d, %u, %u, %u, status)\n", __func__, doLocal, tgoNr, sampleNr, neighNr);
    fflush(stdout);
  }
  if(nlo==NULL ) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }
  if(nlo->totalNr<1 || nlo->xfull==NULL || nlo->_fun==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  if(nlo->totalNr>MAX_PARAMETERS) {
    if(verbose>0) {fprintf(stderr, "Error: too many dimensions to optimize.\n"); fflush(stderr);}
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }

  /* Check if any of the parameters are fixed */
  unsigned int fixedParNr=nloptLimitFixedNr(nlo);
  if(verbose>1 && fixedParNr>0) printf("fixedParNr := %d\n", fixedParNr);
  unsigned int fittedParNr=nlo->totalNr-fixedParNr;
  if(verbose>1) printf("fittedParNr := %d\n", fittedParNr);
  if(fittedParNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }

  /* Check the tolerations */
  for(unsigned int i=0; i<nlo->totalNr; i++) {
    if(nlo->xlower[i]>=nlo->xupper[i]) {nlo->xtol[i]=0.0; continue;}
    if(!(nlo->xtol[i]>0.0)) {
      if(verbose>0) {fprintf(stderr, "Error: invalid xtol[].\n"); fflush(stderr);}
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
      return TPCERROR_NO_DATA;
    }
  }

  /* Continue input checking */
  if(sampleNr<10) sampleNr=50*fittedParNr;
  if(sampleNr&1) sampleNr++; // If number is odd, then add 1
  if(neighNr<2) neighNr=sampleNr/2;
  else if(neighNr>sampleNr-1) neighNr=sampleNr-1; // Make sure "neighNr" isn't too big
  if(tgoNr==0) tgoNr=1+fittedParNr;
  if(verbose>2) {
    printf("sampleNr := %d\n", sampleNr);
    printf("neighNr := %d\n", neighNr);
    printf("tgoNr := %d\n", tgoNr);
    fflush(stdout);
  }

  /* Check if initial guess is valid */
  int initGuessAvailable=1;
  double initGuessCost=nan("");
  for(unsigned int i=0; i<nlo->totalNr; i++) {
    if(isnan(nlo->xfull[i])) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]<nlo->xlower[i]) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]>nlo->xupper[i]) {initGuessAvailable=0; break;}
  }
  if(initGuessAvailable) {
    initGuessCost=nlo->funval=(*nlo->_fun)(nlo->totalNr, nlo->xfull, nlo->fundata);
    nlo->funCalls++;
    if(!isfinite(initGuessCost)) initGuessAvailable=0;
    else if(verbose>2) {
      printf("valid initial guess with cost=%g\n", initGuessCost);
    }
  }

  /* Allocate memory */
  TGO_POINT *points=(TGO_POINT*)malloc(sampleNr*sizeof(TGO_POINT));
  if(points==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
    return TPCERROR_OUT_OF_MEMORY;
  }
  for(unsigned int i=0; i<sampleNr; i++) {
    points[i].topomin=0;
    points[i].fvalue=0.0;
  }

  /*
   *  Iterative TGO, or non-iterative if tgoNr==1
   */
  unsigned int topoNr=0;
  for(unsigned int tgoi=0; tgoi<tgoNr; tgoi++) {

    if(verbose>3 && tgoNr>1) {printf("TGO iteration # %d: \n", 1+tgoi); fflush(stdout);}

    /* Sample N points in the feasible region and compute the object function values for 
       those points which do not already have it. */
    unsigned int badNr=0;
    for(unsigned int si=0; si<sampleNr; si++) if(points[si].topomin==0) {
      if(tgoi==0 && si==0 && initGuessAvailable) {
        /* If initial guess was given, then use it as the first point */
        for(unsigned int i=0; i<nlo->totalNr; i++) points[si].par[i]=nlo->xfull[i];
        points[si].fvalue=initGuessCost;
        continue;
      }
      nloptRandomPoint(points[si].par, nlo->xlower, nlo->xupper, nlo->totalNr);
      points[si].fvalue=(*nlo->_fun)(nlo->totalNr, points[si].par, nlo->fundata);
      nlo->funCalls++;
      /* If function return value was not valid, then we'll try twice with new guesses */
      int tries=0;
      while(!isfinite(points[si].fvalue) && tries<2) {
        if(verbose>8) {
          printf("this point did not give normal return value:\n");
          for(unsigned int i=0; i<nlo->totalNr; i++) printf("  %10.2e", points[si].par[i]);
          printf("\n");
        }
        nloptRandomPoint(points[si].par, nlo->xlower, nlo->xupper, nlo->totalNr);
        points[si].fvalue=(*nlo->_fun)(nlo->totalNr, points[si].par, nlo->fundata);
        nlo->funCalls++;
        tries++;
      }
      if(!isfinite(points[si].fvalue)) badNr++;
    }
    if(verbose>4 && badNr>0) printf("Number of bad points: %d\n", badNr);
    /* Object functions values must be good for at least NeighNr points */
    if(tgoi==0 && (sampleNr-badNr)<=neighNr) {
      if(verbose>0) {
        fprintf(stderr, "Error: invalid values inside parameter range.\n"); fflush(stderr);}
      free(points);
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_BAD_FIT);
      return TPCERROR_BAD_FIT;
    }
    /* Print sampled points */
    if(verbose>10) {
      printf("Sampled points:\n");
      for(unsigned int si=0; si<sampleNr; si++) {
        printf("%d\t", 1+si);
        for(unsigned int i=0; i<nlo->totalNr; i++) printf("%e ", points[si].par[i]);
        printf("=> %e\n", points[si].fvalue);
      }
      fflush(stdout);
    }

    /* For each point i, find out if it is a "topographic minimum", that is, 
       better than k neighbour points. */
    topoNr=0;
    for(unsigned int si=0; si<sampleNr; si++) {
      points[si].topomin=0;

      /* Compute the distances, scaled using xtol[] */
      double sdist[sampleNr];
      for(unsigned int sj=0; sj<sampleNr; sj++) {
        if(si==sj) {sdist[sj]=1.0E+99; continue;}
        for(unsigned int k=0; k<nlo->totalNr; k++) {
          sdist[sj]=0.0;
          if(!(nlo->xtol[k]>0.0)) continue;
          double d=(points[si].par[k]-points[sj].par[k])/nlo->xtol[k];
          if(isfinite(d)) sdist[sj]+=d*d;
        }
        /* Distance should be computed as square root, but it does not affect
           the order of distances; since sqrt() is slow, it is not done. */
        // sdist[sj]=sqrt(sdist[sj]);
      }

      /* Find the closest neighbours for sample i */
      unsigned int neighs[neighNr]; // collect here the neighbours
      unsigned int ni=0;
      for(ni=0; ni<neighNr; ni++) {
        unsigned int mini=0;
        double minv=sdist[mini];
        for(unsigned int sj=0; sj<sampleNr; sj++) {
          if(sdist[sj]<minv) {minv=sdist[sj]; mini=sj;}
        }
        sdist[mini]=1.0E+99; // not used again
        /* If point i is worse than any of the closest neighbours, then point i 
           is certainly not a topographic minimum; then stop this loop and go to
           the next point i+1 */
        if(!(points[si].fvalue<points[mini].fvalue)) break;
        neighs[ni]=mini;
      }
      if(ni<neighNr) continue; // continue with the next i
      /* otherwise mark this as topographic minimum (TM) */
      points[si].topomin=1; topoNr++;
      /* Print TM point */
      if(verbose>6) {
        printf("TM point %d:\n", topoNr);
        printf("%d\t", 1+si);
        for(unsigned int i=0; i<nlo->totalNr; i++) printf("%e ", points[si].par[i]);
        printf("=> %e\n", points[si].fvalue);
        if(verbose>7) {
          printf("  points neighbours:");
          for(ni=0; ni<neighNr; ni++) printf(" %d", 1+neighs[ni]);
          printf("\n");
        }
        fflush(stdout);
      }

    }
    if(verbose>4) {printf("  %d topographical minima\n", topoNr); fflush(stdout);}
    if(topoNr==0) {
      if(verbose>0) {fprintf(stderr, "Warning: no topographical minima found\n"); fflush(stderr);}
      continue;
    }

  } // end of iTGO


  /* If requested, do local optimization for the TMs */
  if(topoNr==0) {
    if(verbose>0) {fprintf(stderr, "Error: no topographical minima found\n"); fflush(stderr);}
    free(points);
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_BAD_FIT);
    return TPCERROR_BAD_FIT;
  }
  for(unsigned int si=0; si<sampleNr; si++) if(points[si].topomin>0) {
    if(verbose>2) printf("LO for TM point %d\n", 1+si);
    NLOPT lnlo; nloptInit(&lnlo);
    nloptDuplicate(nlo, &lnlo);
    for(unsigned int i=0; i<nlo->totalNr; i++) lnlo.xfull[i]=points[si].par[i];
    lnlo.funval=points[si].fvalue;
    for(unsigned int i=0; i<nlo->totalNr; i++) lnlo.xdelta[i]=100.0*lnlo.xtol[i];
    lnlo.maxFunCalls=2000*nlo->totalNr; lnlo.funCalls=0;
    int ret=nloptSimplex(&lnlo, 100*nlo->totalNr, status);
    if(ret==TPCERROR_OK) {
      for(unsigned int i=0; i<nlo->totalNr; i++) points[si].par[i]=lnlo.xfull[i];
      points[si].fvalue=lnlo.funval;
      if(verbose>6) {
        printf("TM point %d after LO:\n", 1+si);
        printf("%d\t", 1+si);
        for(unsigned int i=0; i<nlo->totalNr; i++) printf("%e ", points[si].par[i]);
        printf("=> %e\n", points[si].fvalue); fflush(stdout);
      }
    } else if(verbose>7) {printf("  LO failed\n"); fflush(stdout);}
    nlo->funCalls+=lnlo.funCalls;
    nloptFree(&lnlo);
  }

  /* Find the best minimum and return it */
  {
    unsigned int mini=0;
    double minv=points[mini].fvalue;
    for(unsigned int si=1; si<sampleNr; si++)
      if(!(points[si].fvalue>minv)) {minv=points[si].fvalue; mini=si;}
    for(unsigned int i=0; i<nlo->totalNr; i++) nlo->xfull[i]=points[mini].par[i];
    nlo->funval=points[mini].fvalue;
  }

  free(points);
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Iterative Topographical global optimization algorithm 2 (iTGO2).

    @remark Not well tested!
    @pre Uses rand(), therefore set seed for a new series of pseudo-random numbers; 
    to produce truly random numbers (not just pseudo-random), do srand(time(0)) before 
    calling this function.
    @post The last objective function call is usually not done with the best parameter estimates;
    if objective function simulates data that you need, you must call the function with the final
    parameters.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
    @todo Finalize the code and tests.
 */
int nloptITGO2(
  /** Pointer to NLOPT data. 
      Counter funCalls is initially set to zero, and then increased here.
      Parameter maxFunCalls is used as one of the stopping criteria.
      Parameters xtol[] is used for scaling and as one of the stopping criteria.
   */
  NLOPT *nlo,
  /** Local optimization method: 0=no local optimization, 1=Nelder-Mead. */
  const int doLocal,
  /** Number of TGO iterations; enter 0 to use the default; enter 1 to run TGO
      just once, ie to use TGO instead of iTGO. Iterations are needed if sampleNr
      (below) would otherwise require too much memory. */
  unsigned int tgoNr,
  /** Number of points to sample in one iteration; may need to be large if the number
      of iterations (above) is small; enter 0 to let function decide it. */
  unsigned int sampleNr,
  /** Number of neighbours to investigate (cluster size); enter a large number relative 
      to sampleNr (above) if only few local minima are expected; enter 0 to let
      function decide it. */
  unsigned int neighNr,
  /** Pointer to status data; enter NULL if not needed. */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) {
    printf("%s(NLOPT, %d, %u, %u, %u, status)\n", __func__, doLocal, tgoNr, sampleNr, neighNr);
    fflush(stdout);
  }
  if(nlo==NULL ) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }
  if(nlo->totalNr<1 || nlo->xfull==NULL || nlo->_fun==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  if(nlo->totalNr>MAX_PARAMETERS) {
    if(verbose>0) {fprintf(stderr, "Error: too many dimensions to optimize.\n"); fflush(stderr);}
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }

  /* Check if any of the parameters are fixed */
  unsigned int fixedParNr=nloptLimitFixedNr(nlo);
  if(verbose>1 && fixedParNr>0) printf("fixedParNr := %d\n", fixedParNr);
  unsigned int fittedParNr=nlo->totalNr-fixedParNr;
  if(verbose>1) printf("fittedParNr := %d\n", fittedParNr);
  if(fittedParNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }

  /* Check the tolerations */
  for(unsigned int i=0; i<nlo->totalNr; i++) {
    if(nlo->xlower[i]>=nlo->xupper[i]) {nlo->xtol[i]=0.0; continue;}
    if(!(nlo->xtol[i]>0.0)) {
      if(verbose>0) {fprintf(stderr, "Error: invalid xtol[].\n"); fflush(stderr);}
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
      return TPCERROR_NO_DATA;
    }
  }

  /* Continue input checking */
  if(sampleNr<10) sampleNr=50*fittedParNr;
  if(sampleNr&1) sampleNr++; // If number is odd, then add 1
  if(neighNr<2) neighNr=sampleNr/2;
  else if(neighNr>sampleNr-1) neighNr=sampleNr-1; // Make sure "neighNr" isn't too big
  if(tgoNr==0) tgoNr=1+fittedParNr;
  if(verbose>2) {
    printf("sampleNr := %d\n", sampleNr);
    printf("neighNr := %d\n", neighNr);
    printf("tgoNr := %d\n", tgoNr);
    fflush(stdout);
  }

  /* Check if initial guess is valid */
  int initGuessAvailable=1;
  double initGuessCost=nan("");
  for(unsigned int i=0; i<nlo->totalNr; i++) {
    if(isnan(nlo->xfull[i])) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]<nlo->xlower[i]) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]>nlo->xupper[i]) {initGuessAvailable=0; break;}
  }
  if(initGuessAvailable) {
    initGuessCost=nlo->funval=(*nlo->_fun)(nlo->totalNr, nlo->xfull, nlo->fundata);
    nlo->funCalls++;
    if(!isfinite(initGuessCost)) initGuessAvailable=0;
    else if(verbose>2) {
      printf("valid initial guess with cost=%g\n", initGuessCost);
    }
  }

  /* Allocate memory */
  TGO_POINT *points=(TGO_POINT*)malloc(sampleNr*sizeof(TGO_POINT));
  if(points==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
    return TPCERROR_OUT_OF_MEMORY;
  }
  for(unsigned int i=0; i<sampleNr; i++) {
    points[i].topomin=0;
    points[i].fvalue=0.0;
  }
  TGO_POINT *gpoints=(TGO_POINT*)malloc(sampleNr*sizeof(TGO_POINT));
  if(gpoints==NULL) {
    free(points);
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
    return TPCERROR_OUT_OF_MEMORY;
  }
  for(unsigned int i=0; i<sampleNr; i++) {
    gpoints[i].topomin=0;
    gpoints[i].fvalue=0.0;
  }

  /*
   *  Iterative TGO, or non-iterative if tgoNr==1
   */
  unsigned int topoNr=0;
  if(initGuessAvailable) {
    /* If initial guess was given, then use it as the first G point */
    for(unsigned int i=0; i<nlo->totalNr; i++) gpoints[0].par[i]=nlo->xfull[i];
    gpoints[0].fvalue=initGuessCost;
    topoNr=1;
  }
  for(unsigned int tgoi=0; tgoi<tgoNr; tgoi++) {

    if(verbose>2 && tgoNr>1) {printf("TGO iteration # %d: \n", 1+tgoi); fflush(stdout);}

    while(topoNr<sampleNr) {
      if(verbose>3) {printf("  topoNr := %d\n", topoNr); fflush(stdout);}

      /* Sample N points and compute the object function values for all points. */
      unsigned int badNr=0;
      for(unsigned int si=0; si<sampleNr; si++) {
        points[si].topomin=0;
        nloptRandomPoint(points[si].par, nlo->xlower, nlo->xupper, nlo->totalNr);
        points[si].fvalue=(*nlo->_fun)(nlo->totalNr, points[si].par, nlo->fundata);
        nlo->funCalls++;
        /* If function return value was not valid, then we'll try twice with new guesses */
        int tries=0;
        while(!isfinite(points[si].fvalue) && tries<2) {
          if(verbose>8) {
            printf("this point did not give normal return value:\n");
            for(unsigned int i=0; i<nlo->totalNr; i++) printf("  %10.2e", points[si].par[i]);
            printf("\n");
          }
          nloptRandomPoint(points[si].par, nlo->xlower, nlo->xupper, nlo->totalNr);
          points[si].fvalue=(*nlo->_fun)(nlo->totalNr, points[si].par, nlo->fundata);
          nlo->funCalls++;
          tries++;
        }
        if(!isfinite(points[si].fvalue)) badNr++;
      }
      if(verbose>4 && badNr>0) printf("Number of bad points: %d\n", badNr);
      /* Object functions values must be good for at least NeighNr points */
      if(tgoi==0 && (sampleNr-badNr)<=neighNr) {
        if(verbose>0) {
          fprintf(stderr, "Error: invalid values inside parameter range.\n"); fflush(stderr);}
        free(points); free(gpoints);
        statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_BAD_FIT);
        return TPCERROR_BAD_FIT;
      }

      /* For each point i, find out if it is a "topographic minimum", that is, 
         better than k neighbour points. */
      for(unsigned int si=0; si<sampleNr; si++) {
        points[si].topomin=0;

        /* Compute the distances, scaled using xtol[] */
        double sdist[sampleNr];
        for(unsigned int sj=0; sj<sampleNr; sj++) {
          if(si==sj) {sdist[sj]=1.0E+99; continue;}
          for(unsigned int k=0; k<nlo->totalNr; k++) {
            sdist[sj]=0.0;
            if(!(nlo->xtol[k]>0.0)) continue;
            double d=(points[si].par[k]-points[sj].par[k])/nlo->xtol[k];
            if(isfinite(d)) sdist[sj]+=d*d;
          }
          /* Distance should be computed as square root, but it does not affect
             the order of distances; since sqrt() is slow, it is not done. */
          // sdist[sj]=sqrt(sdist[sj]);
        }

        /* Find the closest neighbours for sample i */
        unsigned int ni=0;
        for(ni=0; ni<neighNr; ni++) {
          unsigned int mini=0;
          double minv=sdist[mini];
          for(unsigned int sj=0; sj<sampleNr; sj++) {
            if(sdist[sj]<minv) {minv=sdist[sj]; mini=sj;}
          }
          sdist[mini]=1.0E+99; // not used again
          /* If point i is worse than any of the closest neighbours, then point i is certainly not 
             a topographic minimum; then stop this loop and go to the next point i+1 */
          if(!(points[si].fvalue<points[mini].fvalue)) break;
        }
        if(ni<neighNr) continue; // continue with the next sample i
        /* otherwise mark this as topographic minimum (TM) */
        points[si].topomin=1;
        /* and copy it to G point list (if there is space left) */
        if(topoNr<sampleNr) {
          for(unsigned int i=0; i<nlo->totalNr; i++) gpoints[topoNr].par[i]=points[si].par[i];
          gpoints[topoNr++].fvalue=points[si].fvalue;
        }
        /* Print TM point */
        if(verbose>6) {
          printf("TM point %d\t", 1+si);
          for(unsigned int i=0; i<nlo->totalNr; i++) printf("%e ", points[si].par[i]);
          printf("=> %e\n", points[si].fvalue);
          fflush(stdout);
        }
      }
    } // inner loop 


    if(verbose>3) {printf("  process gpoints.\n"); fflush(stdout);}
    /* For each point i, find out if it is a "topographic minimum", that is, 
       better than k/2 neighbour points. */
    for(unsigned int si=0; si<sampleNr; si++) {
      gpoints[si].topomin=0;

      /* Compute the distances, scaled using xtol[] */
      double sdist[sampleNr];
      for(unsigned int sj=0; sj<sampleNr; sj++) {
        if(si==sj) {sdist[sj]=1.0E+99; continue;}
        for(unsigned int k=0; k<nlo->totalNr; k++) {
          sdist[sj]=0.0;
          if(!(nlo->xtol[k]>0.0)) continue;
          double d=(gpoints[si].par[k]-gpoints[sj].par[k])/nlo->xtol[k];
          if(isfinite(d)) sdist[sj]+=d*d;
        }
      }

      /* Find the closest neighbours for sample i */
      unsigned int ni=0;
      for(ni=0; ni<neighNr; ni++) {
        unsigned int mini=0;
        double minv=sdist[mini];
        for(unsigned int sj=0; sj<sampleNr; sj++) {
          if(sdist[sj]<minv) {minv=sdist[sj]; mini=sj;}
        }
        sdist[mini]=1.0E+99; // not used again
        /* If point i is worse than any of the closest neighbours, then point i is certainly not 
           a topographic minimum; then stop this loop and go to the next point i+1 */
        if(!(gpoints[si].fvalue<gpoints[mini].fvalue)) break;
      }
      if(ni<neighNr) continue; // continue with the next sample i
      /* otherwise mark this as topographic minimum (TM) */
      gpoints[si].topomin=1;
      /* Print TM point */
      if(verbose>6) {
        printf("GTM point %d\t", 1+si);
        for(unsigned int i=0; i<nlo->totalNr; i++) printf("%e ", gpoints[si].par[i]);
        printf("=> %e\n", gpoints[si].fvalue);
        fflush(stdout);
      }
    }
    /* Copy TMs into the start of gpoint list and set topoNr accordingly */
    topoNr=0;
    for(unsigned int si=0; si<sampleNr; si++) if(gpoints[si].topomin>0) {
      if(si>topoNr) {
        for(unsigned int i=0; i<nlo->totalNr; i++) gpoints[topoNr].par[i]=gpoints[si].par[i];
        gpoints[topoNr].fvalue=gpoints[si].fvalue;
      }
      topoNr++;
    }
    if(verbose>3) {printf("  %d topographical minima\n", topoNr); fflush(stdout);}
    if(topoNr==0) {
      if(verbose>0) {fprintf(stderr, "Warning: no topographical minima found\n"); fflush(stderr);}
      continue;
    }


  } // end of iTGO


  /* If requested, do local optimization for the TMs */
  if(topoNr==0) {
    if(verbose>0) {fprintf(stderr, "Error: no topographical minima found\n"); fflush(stderr);}
    free(points); free(gpoints);
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_BAD_FIT);
    return TPCERROR_BAD_FIT;
  }
  for(unsigned int si=0; si<sampleNr; si++) if(gpoints[si].topomin>0) {
    if(verbose>2) printf("LO for TM point %d\n", 1+si);
    if(verbose>3) {
      printf("TM point %d before LO:\n", 1+si);
      printf("%d\t", 1+si);
      for(unsigned int i=0; i<nlo->totalNr; i++) printf("%e ", gpoints[si].par[i]);
      printf("=> %e\n", gpoints[si].fvalue); fflush(stdout);
    }
    NLOPT lnlo; nloptInit(&lnlo);
    nloptDuplicate(nlo, &lnlo);
    for(unsigned int i=0; i<nlo->totalNr; i++) lnlo.xfull[i]=gpoints[si].par[i];
    lnlo.funval=points[si].fvalue;
    for(unsigned int i=0; i<nlo->totalNr; i++) lnlo.xdelta[i]=100.0*lnlo.xtol[i];
    lnlo.maxFunCalls=2000*nlo->totalNr; lnlo.funCalls=0;
    int ret=nloptSimplex(&lnlo, 100*nlo->totalNr, status);
    if(ret==TPCERROR_OK) {
      for(unsigned int i=0; i<nlo->totalNr; i++) gpoints[si].par[i]=lnlo.xfull[i];
      gpoints[si].fvalue=lnlo.funval;
      if(verbose>3) {
        printf("TM point %d after LO:\n", 1+si);
        printf("%d\t", 1+si);
        for(unsigned int i=0; i<nlo->totalNr; i++) printf("%e ", gpoints[si].par[i]);
        printf("=> %e\n", gpoints[si].fvalue); fflush(stdout);
      }
    } else if(verbose>7) {printf("  LO failed\n"); fflush(stdout);}
    nlo->funCalls+=lnlo.funCalls;
    nloptFree(&lnlo);
  }

  /* Find the best minimum and return it */
  {
    unsigned int mini=0;
    double minv=gpoints[mini].fvalue;
    for(unsigned int si=1; si<sampleNr; si++)
      if(!(gpoints[si].fvalue>minv)) {minv=gpoints[si].fvalue; mini=si;}
    for(unsigned int i=0; i<nlo->totalNr; i++) nlo->xfull[i]=gpoints[mini].par[i];
    nlo->funval=gpoints[mini].fvalue;
  }

  free(points); free(gpoints);
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Iterative accumulative topographical global optimization algorithm (ITGO).

    This function is aimed for problems where objective function call is computationally
    slow compared to processing an increasing list of already sampled points.

    @remark Not finalized!
    @pre Uses rand(), therefore set seed for a new series of pseudo-random numbers; 
    to produce truly random numbers (not just pseudo-random), do srand(time(0)) before 
    calling this function.
    @post The last objective function call is usually not done with the best parameter estimates;
    if objective function simulates data that you need, you must call the function with the final
    parameters.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
    @todo Finalize the code and tests.
 */
int nloptIATGO(
  /** Pointer to NLOPT data. 
      Counter funCalls is initially set to zero, and then increased here.
      Parameter maxFunCalls is used as one of the stopping criteria.
      Parameters xtol[] is used for scaling and as one of the stopping criteria.
   */
  NLOPT *nlo,
  /** Local optimization method: 0=no local optimization, 1=Nelder-Mead. */
  const int doLocal,
  /** Number of TGO iterations; enter 0 to use the default; enter 1 to run TGO
      just once, ie to use TGO instead of iTGO. Iterations are needed if sampleNr
      (below) would otherwise require too much memory. */
  unsigned int maxIterNr,
  /** Fraction of samples to define as neighbours (cluster size); enter a large fraction (>0.5) 
      if only few local minima are expected. */
  double neighFract,
  /** Pointer to status data; enter NULL if not needed. */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) {
    printf("%s(NLOPT, %d, %u, %g, status)\n", __func__, doLocal, maxIterNr, neighFract);
    fflush(stdout);
  }
  if(nlo==NULL ) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }
  if(nlo->totalNr<1 || nlo->xfull==NULL || nlo->_fun==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }

  /* Check if any of the parameters are fixed */
  unsigned int fixedParNr=nloptLimitFixedNr(nlo);
  if(verbose>1 && fixedParNr>0) printf("fixedParNr := %d\n", fixedParNr);
  unsigned int fittedParNr=nlo->totalNr-fixedParNr;
  if(verbose>1) printf("fittedParNr := %d\n", fittedParNr);
  if(fittedParNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }

  /* Check the tolerations */
  for(unsigned int i=0; i<nlo->totalNr; i++) {
    if(nlo->xlower[i]>=nlo->xupper[i]) {nlo->xtol[i]=0.0; continue;}
    if(!(nlo->xtol[i]>0.0)) {
      if(verbose>0) {fprintf(stderr, "Error: invalid xtol[].\n"); fflush(stderr);}
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
      return TPCERROR_NO_DATA;
    }
  }

  /* Continue input checking */
  unsigned int sampleNr=100*fittedParNr;
  if(neighFract<0.04) neighFract=0.04; else if(neighFract>0.97) neighFract=0.97;
  if(maxIterNr==0) maxIterNr=1+2*fittedParNr;
  if(verbose>2) {
    printf("sampleNr := %d\n", sampleNr);
    printf("neighFract := %g\n", neighFract);
    printf("maxIterNr := %d\n", maxIterNr);
    fflush(stdout);
  }

  /* Check if initial guess is valid */
  int initGuessAvailable=1;
  double initGuessCost=nan("");
  for(unsigned int i=0; i<nlo->totalNr; i++) {
    if(isnan(nlo->xfull[i])) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]<nlo->xlower[i]) {initGuessAvailable=0; break;}
    if(nlo->xfull[i]>nlo->xupper[i]) {initGuessAvailable=0; break;}
  }
  if(initGuessAvailable) {
    initGuessCost=nlo->funval=(*nlo->_fun)(nlo->totalNr, nlo->xfull, nlo->fundata);
    nlo->funCalls++;
    if(!isfinite(initGuessCost)) initGuessAvailable=0;
    else if(verbose>2) {
      printf("valid initial guess with cost=%g\n", initGuessCost);
    }
  }

  /*
   *  Initialize the sample list with random points
   */

  /* Allocate memory */
  unsigned int dim=nlo->totalNr;
  nlo->usePList=1;
  nlo->plist=(double*)malloc(sampleNr*(dim+1)*sizeof(double));
  if(nlo->plist==NULL) { // will be freed with nloptFree()
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
    return TPCERROR_OUT_OF_MEMORY;
  }
  unsigned int *tmlist=(unsigned int*)malloc(sampleNr*sizeof(double));
  if(tmlist==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
    return TPCERROR_OUT_OF_MEMORY;
  }
  for(unsigned int i=0; i<sampleNr; i++) tmlist[i]=0;

  /* Fill the list with random points and compute the object function values */
  unsigned int badNr=0;
  for(unsigned int si=0; si<sampleNr; si++) {
    if(si==0 && initGuessAvailable) {
      /* If initial guess was given, then use it as the first point */
      for(unsigned int i=0; i<nlo->totalNr; i++) nlo->plist[si+i]=nlo->xfull[i];
      nlo->plist[si+dim]=initGuessCost;
      continue;
    }
    nloptRandomPoint(&nlo->plist[si*(dim+1)], nlo->xlower, nlo->xupper, dim);
    nlo->plist[si*(dim+1)+dim]=(*nlo->_fun)(nlo->totalNr, &nlo->plist[si*(dim+1)], nlo->fundata);
    nlo->funCalls++;
    if(!isfinite(nlo->plist[si*(dim+1)+dim])) badNr++;
  }
  if(verbose>4 && badNr>0) printf("Number of bad points: %d\n", badNr);
  /* Not all object function values can be bad */
  if((sampleNr-badNr)<10) {
    if(verbose>0) {
      fprintf(stderr, "Error: invalid values inside parameter range.\n"); fflush(stderr);}
    free(tmlist);
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_BAD_FIT);
    return TPCERROR_BAD_FIT;
  }
  /* Print sampled points */
  if(verbose>100) {
    printf("Sampled points:\n");
    for(unsigned int si=0; si<sampleNr; si++) {
      printf("%d\t", 1+si);
      for(unsigned int i=0; i<dim; i++) printf("%e ", nlo->plist[si*(dim+1)+i]);
      printf("=> %e\n", nlo->plist[si*(dim+1)+dim]);
    }
    fflush(stdout);
  }
  

  /*
   *  Start iterations
   */
  unsigned int neighNr=neighFract*sampleNr;
  unsigned int iterNr=0;
  while(iterNr<maxIterNr && nlo->funCalls<nlo->maxFunCalls) {
    iterNr++;
    if(verbose>2) {printf("iteration %d with %d samples\n", iterNr, sampleNr); fflush(stdout);}
    if(verbose>3) {printf("  sampleNr=%d funCalls=%d\n", sampleNr, nlo->funCalls); fflush(stdout);}

    if(verbose>10) {
      printf("Sampled points:\n");
      for(unsigned int si=0; si<sampleNr; si++) {
        printf("%d\t", 1+si);
        for(unsigned int i=0; i<dim; i++) printf("%e ", nlo->plist[si*(dim+1)+i]);
        printf("=> %e\n", nlo->plist[si*(dim+1)+dim]);
      }
      fflush(stdout);
    }

    /* Find topographic minima (TM) */
    unsigned int topoNr=0;
    if(verbose>3) {printf("neighNr := %d\n", neighNr); fflush(stdout);}

    /* For each point i, find out if it is a "topographic minimum", that is, 
       better than k neighbour points. */
    for(unsigned int si=0; si<sampleNr; si++) {
      tmlist[si]=0;
      
      /* If function value is invalid, then this cannot be accepted as TM */
      if(!isfinite(nlo->plist[si*(dim+1)+dim])) continue;

      /* Compute the distances, scaled using xtol[] */
      double sdist[sampleNr];
      for(unsigned int sj=0; sj<sampleNr; sj++) {
        if(si==sj) {sdist[sj]=1.0E+99; continue;}
        for(unsigned int k=0; k<dim; k++) {
          sdist[sj]=0.0;
          if(!(nlo->xtol[k]>0.0)) continue;
          double d=(nlo->plist[si*(dim+1)+k]-nlo->plist[sj*(dim+1)+k])/nlo->xtol[k];
          if(isfinite(d)) sdist[sj]+=d*d;
        }
      }

      /* Find the closest neighbours for sample i */
      unsigned int ni=0;
      for(ni=0; ni<neighNr; ni++) {
        unsigned int mini=0;
        double minv=sdist[mini];
        for(unsigned int sj=0; sj<sampleNr; sj++) {
          if(sdist[sj]<minv) {minv=sdist[sj]; mini=sj;}
        }
        sdist[mini]=1.0E+99; // not used again
        /* If point i is worse than any of the closest neighbours, then point i is certainly not 
           a topographic minimum; then stop this loop and go to the next point i+1 */
        if(nlo->plist[si*(dim+1)+dim] > nlo->plist[mini*(dim+1)+dim]) break;
      }
      if(ni<neighNr) continue; // continue with the next sample i
      /* otherwise mark this as topographic minimum (TM) */
      tmlist[si]=1;
      topoNr++;
      /* Print TM point */
      if(verbose>6) {
        printf("TM point %d\t", 1+si);
        for(unsigned int i=0; i<dim; i++) printf("%e ", nlo->plist[si*(dim+1)+i]);
        printf("=> %e\n", nlo->plist[si*(dim+1)+dim]);
        fflush(stdout);
      }
    }
    if(verbose>2) {printf("topographic minima: %d\n", topoNr); fflush(stdout);}

    /* Local optimization from each TM */
    if(topoNr==0) {
      if(verbose>0) {fprintf(stderr, "Error: no topographical minima found\n"); fflush(stderr);}
      free(tmlist);
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_BAD_FIT);
      return TPCERROR_BAD_FIT;
    }
    for(unsigned int si=0; si<sampleNr; si++) if(tmlist[si]>0) {
      if(verbose>2) printf("LO for TM point %d\n", 1+si);
      if(verbose>3) {
        printf("TM point %d before LO:\n", 1+si);
        printf("%d\t", 1+si);
        for(unsigned int i=0; i<dim; i++) printf("%e ", nlo->plist[si*(dim+1)+i]);
        printf("=> %e\n", nlo->plist[si*(dim+1)+dim]); fflush(stdout);
      }
      NLOPT lnlo; nloptInit(&lnlo);
      nloptDuplicate(nlo, &lnlo);
      for(unsigned int i=0; i<dim; i++) lnlo.xfull[i]=nlo->plist[si*(dim+1)+i];
      lnlo.funval=nlo->plist[si*(dim+1)+dim];
      for(unsigned int i=0; i<dim; i++) lnlo.xdelta[i]=100.0*lnlo.xtol[i];
      lnlo.maxFunCalls=1000*dim; lnlo.funCalls=0;
      int ret=nloptSimplex(&lnlo, 100*dim, NULL);
      if(ret==TPCERROR_OK) {
        if(verbose>3) {
          printf("TM point %d after LO:\n", 1+si);
          printf("%d\t", 1+si);
          for(unsigned int i=0; i<dim; i++) printf("%e ", lnlo.xfull[i]);
          printf("=> %e\n", lnlo.funval); fflush(stdout);
        }
        if(verbose>5) printf("localfunCalls=%d\n", lnlo.funCalls);
      } else if(verbose>7) {printf("  LO failed\n"); fflush(stdout);}
      /* Add to list the points visited by local optimization */
      for(unsigned int i=0; i<lnlo.funCalls; i++) {
        nlo->funCalls++;
        int ret=nloptAddP(nlo, &lnlo.plist[i*(dim+1)], lnlo.plist[i*(dim+1)+dim]);
        if(ret!=TPCERROR_OK) {
          free(tmlist);
          statusSet(status, __func__, __FILE__, __LINE__, ret);
          return(ret);
        }
      }

//        for(unsigned int i=0; i<dim; i++) nlo->plist[si*(dim+1)+i]=lnlo.xfull[i];
//        nlo->plist[si*(dim+1)+dim]=lnlo.funval;
//      nlo->funCalls+=lnlo.funCalls;
      nloptFree(&lnlo);
    }
    sampleNr=nlo->funCalls;
    /* Reallocate also tmlist */
    tmlist=(unsigned int*)realloc(tmlist, sampleNr*sizeof(unsigned int));
    if(tmlist==NULL) {
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
      return TPCERROR_OUT_OF_MEMORY;
    }
    if(verbose>7) printf("   iteration end, sampleNr=%d\n", sampleNr);
  } // end of iterations


  /* Find the best minimum and return it */
  {
    unsigned int mini=0;
    double minv=nlo->plist[dim];
    for(unsigned int si=1; si<sampleNr; si++)
      if(!(nlo->plist[si*(dim+1)+dim]>minv)) {minv=nlo->plist[si*(dim+1)+dim]; mini=si;}
    for(unsigned int i=0; i<dim; i++) nlo->xfull[i]=nlo->plist[mini*(dim+1)+i];
    nlo->funval=nlo->plist[mini*(dim+1)+dim];
  }

printf("neighNr=%d\n", neighNr);

  free(tmlist);
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
