/** @file tacorder.c
 *  @brief Sort or otherwise change the order of data in TAC struct.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
//#include "tpcift.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpctac.h"
/*****************************************************************************/

/*****************************************************************************/
/** Verify that x values (sample or frame times) in TAC struct are ordered by
    increasing x. 
    Overlapping samples are ok here.

    @return enum tpcerror (TPCERROR_OK when in increasing order).
    @return 0 if in increasing order, >0 otherwise or in case of error.
    @author Vesa Oikonen
    @sa tacSortByTime, tacDeleteFrameOverlap.
 */
int tacVerifyTimeOrder(
  /** Pointer to TAC struct */
  TAC *d,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  /* Check that required data exists */
  if(d==NULL || d->sampleNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  /* All sample times must be available */
  if(tacXNaNs(d)>0) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_MISSING_DATA);
    return TPCERROR_MISSING_DATA;
  }
  /* Less than two samples are always in order */
  if(d->sampleNr<2) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
    return TPCERROR_OK;
  }

  int fi;
  double p, s;
  
  fi=0; if(d->isframe==0) p=d->x[fi]; else p=0.5*(d->x1[fi]+d->x2[fi]);
  for(fi=1; fi<d->sampleNr; fi++) {
    if(d->isframe==0) s=d->x[fi]; else s=0.5*(d->x1[fi]+d->x2[fi]);
    if(s<p) {
      if(verbose>2) printf("x[%d]=%g\nx[%d]=%g\n", fi-1, p, fi, s);
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OVERLAPPING_DATA);
      return(TPCERROR_OVERLAPPING_DATA);
    }
    p=s;
  }
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Sort samples (frames) in TAC struct by increasing sample time.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
    @sa tacVerifyTimeOrder, tacSortByConc, tacSortByAUC, tacDeleteFrameOverlap
 */
int tacSortByTime(
  /** Pointer to TAC struct */
  TAC *d,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  /* Check that required data exists */
  if(d==NULL || d->tacNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  /* All sample times must be available */
  if(tacXNaNs(d)>0) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_MISSING_DATA);
    return TPCERROR_MISSING_DATA;
  }
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  if(d->sampleNr<1) return TPCERROR_OK;
  
  TACC *c;
  int ri, fi, fj;
  double s, a1, a2;
  for(fi=0; fi<d->sampleNr-1; fi++) for(fj=fi+1; fj<d->sampleNr; fj++) {
    if(d->isframe==0) {a1=d->x[fi]; a2=d->x[fj];}
    else {a1=0.5*(d->x1[fi]+d->x2[fi]); a2=0.5*(d->x1[fj]+d->x2[fj]);}
    if(a2>=a1) continue;
    s=d->x[fi];  d->x[fi]=d->x[fj];   d->x[fj]=s;
    s=d->x1[fi]; d->x1[fi]=d->x1[fj]; d->x1[fj]=s;
    s=d->x2[fi]; d->x2[fi]=d->x2[fj]; d->x2[fj]=s;
    s=d->w[fi];  d->w[fi]=d->w[fj];   d->w[fj]=s;
    for(ri=0; ri<d->tacNr; ri++) {
      c=d->c+ri;
      s=c->y[fi];  c->y[fi]=c->y[fj];   c->y[fj]=s;
    }
  }  
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Sort samples (frames) in TAC struct by decreasing sample value.

    @sa tacSortByTime
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
 */
int tacSortByConc(
  /** Pointer to TAC struct */
  TAC *d,
  /** Index of y column which the sorting is based on */
  const int i,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  /* Check that required data exists */
  if(d==NULL || d->tacNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  if(i<0 || i>=d->_tacNr) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  if(d->sampleNr<1) return TPCERROR_OK;

  TACC *c;
  int ri, fi, fj;
  double s, a1, a2;
  for(fi=0; fi<d->sampleNr-1; fi++) for(fj=fi+1; fj<d->sampleNr; fj++) {
    a1=d->c[i].y[fi]; a2=d->c[i].y[fj];
    if(a2<=a1) continue;
    s=d->x[fi];  d->x[fi]=d->x[fj];   d->x[fj]=s;
    s=d->x1[fi]; d->x1[fi]=d->x1[fj]; d->x1[fj]=s;
    s=d->x2[fi]; d->x2[fi]=d->x2[fj]; d->x2[fj]=s;
    s=d->w[fi];  d->w[fi]=d->w[fj];   d->w[fj]=s;
    for(ri=0; ri<d->_tacNr; ri++) {
      c=d->c+ri;
      s=c->y[fi];  c->y[fi]=c->y[fj];   c->y[fj]=s;
    }
  }
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/// @cond
/** Local function */
static int tacQSortName(const void *c1, const void *c2)
{
  int res;
  res=strcasecmp( ((TACC*)c1)->name, ((TACC*)c2)->name );
  return(res);
}
/// @endcond
/** Sort TACs in alphabetical order by their TAC name.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
    @sa tacSortByAUC, tacSortByTime
 */
int tacSortByName(
  /** Pointer to TAC struct */
  TAC *d,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  /* Check that required data exists */
  if(d==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  if(d->tacNr<2) return TPCERROR_OK;
  qsort(d->c, d->tacNr, sizeof(TACC), tacQSortName);
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Sort TACs in decreasing order by their area-under-curve (AUC).

    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
    @sa tacSortByName, tacSortByTime
 */
int tacSortByAUC(
  /** Pointer to TAC struct */
  TAC *d,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  /* Check that required data exists */
  if(d==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  if(d->tacNr<2) return TPCERROR_OK;
  /* Calculate AUCs for each TAC */
  double auc[d->tacNr], fdur, v;
  int i, j;
  for(i=0; i<d->tacNr; i++) {
    auc[i]=0.0;
    for(j=0; j<d->sampleNr; j++) {
      fdur=d->x2[0]-d->x1[0]; v=d->c[i].y[j]; 
      if(isnan(fdur) || isnan(v)) continue;
      auc[i]+=fdur*v;
    }
  }
  /* Sort */
  double a;
  for(i=0; i<d->tacNr-1; i++)
    for(j=i+1; j<d->tacNr; j++)
      if(auc[j]>auc[i]) {
        tacSwapTACCs(d, j, i); 
        a=auc[i]; auc[i]=auc[j]; auc[j]=a;
      }
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Move TACC from one position to another inside TAC struct, moving the
 *  positions of other TACCs accordingly.
 *  @return enum tpcerror (TPCERROR_OK when successful).
 *  @author Vesa Oikonen
 */
int tacMoveTACC(
  /** Pointer to TAC data. Main header and tacNr are not changed. */
  TAC *d,
  /** TACC index [0.._tacNr-1] */
  int from,
  /** TACC index [0.._tacNr-1] */
  int to
) {
  int i;
  TACC tacc;
  size_t taccSize;

  if(d==NULL || from<0 || to<0) return(TPCERROR_FAIL);
  if(from>=d->_tacNr || to>=d->_tacNr) return(TPCERROR_FAIL);
  if(from==to) return(TPCERROR_OK); // nothing to do
  taccSize=sizeof(TACC);
  memcpy(&tacc, d->c+from, taccSize);
  if(from>to) for(i=from; i>to; i--)
    memcpy(d->c+i, d->c+(i-1), taccSize);
  else for(i=from; i<to; i++)
    memcpy(d->c+i, d->c+(i+1), taccSize);
  memcpy(d->c+i, &tacc, taccSize);
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Change places between two TACCs inside TAC struct.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
 */
int tacSwapTACCs(
  /** Pointer to TAC data. Main header and tacNr are not changed. */
  TAC *d,
  /** TACC index 1 [0.._tacNr-1] */
  int i1,
  /** TACC index 2 [0.._tacNr-1] */
  int i2
) {
  TACC tacc;
  size_t taccSize;

  if(d==NULL || i1<0 || i2<0) return(TPCERROR_FAIL);
  if(i1>=d->_tacNr || i2>=d->_tacNr) return(TPCERROR_FAIL);
  if(i1==i2) return(TPCERROR_OK); // nothing to do
  taccSize=sizeof(TACC);
  memcpy(&tacc, d->c+i1, taccSize);
  memcpy(d->c+i1, d->c+i2, taccSize);
  memcpy(d->c+i2, &tacc, taccSize);
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Delete specified TACC inside TAC struct, moving the
    positions of other TACCs accordinly.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
 */
int tacDeleteTACC(
  /** Pointer to TAC data. Main header is not changed, but tacNr is decreased.*/
  TAC *d,
  /** TACC index [0..tacNr-1] */
  int i
) {
  /* Check that region exists */
  if(d==NULL || i<0 || i>=d->tacNr) return(TPCERROR_FAIL);
  /* If the last TACC is to be deleted, then just decrease the tacNr */
  if(i==d->tacNr-1) {d->tacNr--; return(TPCERROR_OK);}
  /* Otherwise move it to the last position, and then decrease tacNr */
  if(tacMoveTACC(d, i, d->tacNr-1)!=TPCERROR_OK) return(TPCERROR_FAIL);
  d->tacNr--;
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
