/** @file fit_dexp.c
    @brief Fit the sum of exponentials to decaying TAC.
    @copyright (c) Turku PET Centre
    @author Vesa Oikonen
 */
/// @cond
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
#include "tpcextensions.h"
#include "tpcift.h"
#include "tpctac.h"
#include "tpcpar.h"
#include "tpcbfm.h"
#include "tpctacmod.h"
#include "tpclinopt.h"
#include "tpcrand.h"
#include "tpcnlopt.h"
/*****************************************************************************/

/*****************************************************************************/
/* Local functions */
double func_exp(int parNr, double *p, void*);
/*****************************************************************************/
typedef struct FITDATA {
  /** Nr of TAC samples */
  unsigned int n;
  /** Array of x values */
  double       *x;
  /** Array of measured y values */
  double       *ymeas;
  /** Array for simulated y values */
  double       *ysim;
  /** Array of weight values */
  double       *w;
} FITDATA;
/*****************************************************************************/

/*****************************************************************************/
static char *info[] = {
  "Non-linear fitting of the sum of 1-5 exponential functions to decaying",
  "time-activity curve (TAC).",
  " ",
  "Function:",
  " ",
  "  f(x) = a1*exp(k1*x) + a2*exp(k2*x) + a3*exp(k3*x) + ...",
  " ",
  ", where a's > 0 and k's < 0.",
  " ",
  "Usage: @P [Options] tacfile parfile",
  " ",
  "Options:",
  " -5 | -4 | -3 | -2 | -1",
  "     Fit to sum of specified number of exponential functions; by default",
  "     determined automatically.",
  " -w1",
  "     All weights are set to 1.0 (no weighting); by default, weights in",
  "     data file are used, if available.",
  " -wf",
  "     Weight by sampling interval.",
  " -svg=<Filename>",
  "     Fitted and measured TACs are plotted in specified SVG file.",
  " -nr=<value>",
  "     Number of basis functions, used in determination of initial parameter",
  "     estimates and number of exponentials for nonlinear fitting;",
  "     default is 100.",
  " ",
  " -stdoptions", // List standard options like --help, -v, etc
  " ",
  "If TAC file contains more than one TAC, only the first is used.",
  "Function parameters are written in fit format.",
  " ",
  "See also: fit_exp, fit_fexp, fit_ratf, fit2dat, extrapol, paucinf, tacln",
  " ",
  "Keywords: TAC, curve fitting, extrapolation",
  0};
/*****************************************************************************/

/*****************************************************************************/
/* Turn on the globbing of the command line, since it is disabled by default in
   mingw-w64 (_dowildcard=0); in MinGW32 define _CRT_glob instead, if necessary;
   In Unix&Linux wildcard command line processing is enabled by default. */
/*
#undef _CRT_glob
#define _CRT_glob -1
*/
int _dowildcard = -1;
/*****************************************************************************/

/*****************************************************************************/
/// @endcond
/** Spectral x,y-data fitting to the sum of decaying exponential functions.
    @details f(x) = a1*exp(-k1*x) + a2*exp(-k2*x) + a3*exp(-k3*x) + ...
     where aN>=0 and kN>0.
    @todo Add tests.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
 */
int spectralDExp(
  /** Pointer to TAC x data (not modified). Data must be sorted by increasing x.
      Negative or missing x values are not allowed. */
  const double *x,
  /** Pointer to TAC y data (not modified). Data must be sorted by increasing x.
      Missing y values are not allowed. */
  const double *y,
  /** Pointer to TAC sample weights (not modified). Enter NULL if not needed. */
  double *w,
  /** Number of samples in x[] and y[]; at least 3. */
  const int snr,
  /** Minimum of k (set based on sample range and unit); must be >0. */
  const double kmin,
  /** Maximum of k (set based on sample range and unit); must be >kmin. */
  const double kmax,
  /** Number of basis functions to calculate; 
      also the length of arrays k[] and a[]; at least 4. */
  const int bnr,
  /** Pointer to array of length bnr for k values, filled in by this function. */
  double *k,
  /** Pointer to array of length bnr for a values, filled in by this function;
      notice that most values may be set to zero. */
  double *a,
  /** Pointer to array for fitted y values. Enter NULL if not needed. */
  double *yfit,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  if(x==NULL || y==NULL || k==NULL || a==NULL || snr<1 || bnr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  if(snr<3 || bnr<4) {
    if(verbose>1) fprintf(stderr, "invalid sample or BF number\n");
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_TOO_FEW);
    return TPCERROR_TOO_FEW;
  }
  if(!(kmin>0.0) || kmin>=kmax) { // range cannot be computed otherwise
    if(verbose>1) fprintf(stderr, "invalid k range\n");
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_VALUE);
    return TPCERROR_INVALID_VALUE;
  }
  int ret;

  if(verbose>1) printf("computing k values\n");
  {
    double a, b, c;
    a=log10(kmin); b=log10(kmax); c=(b-a)/(double)(bnr-1);
    if(verbose>3) printf(" a := %g\n b := %g\n c := %g\n", a, b, c);
    for(int bi=0; bi<bnr; bi++) k[bi]=pow(10.0, (double)bi*c+a);
  }

  /* Allocate memory required by NNLS */
  int     nnls_n, nnls_m, n, m, nnls_index[bnr];
  double *nnls_a[bnr], *nnls_b, *nnls_zz, nnls_x[bnr], *nnls_mat,
          nnls_wp[bnr], *dptr, nnls_rnorm;
  if(verbose>1) printf("allocating memory for NNLS\n");
  nnls_n=bnr; nnls_m=snr;
  nnls_mat=(double*)malloc(((nnls_n+2)*nnls_m)*sizeof(double));
  if(nnls_mat==NULL) {
    if(verbose>1) fprintf(stderr, "Error: cannot allocate memory for NNLS.\n");
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OUT_OF_MEMORY);
    return TPCERROR_OUT_OF_MEMORY;
  }
  for(n=0, dptr=nnls_mat; n<nnls_n; n++) {nnls_a[n]=dptr; dptr+=nnls_m;}
  nnls_b=dptr; dptr+=nnls_m; nnls_zz=dptr;

  if(verbose>1) printf("filling NNLS data matrix\n");
  /* Fill NNLS B array with measured y values */
  for(m=0; m<nnls_m; m++) nnls_b[m]=y[m];

  /* Fill NNLS A matrix with basis functions, calculated in place */
  for(n=0; n<nnls_n; n++)
    for(m=0; m<nnls_m; m++)
      nnls_a[n][m]=exp(-k[n]*x[m]);

  /* Apply weights, if given */
  if(w!=NULL) nnlsWght(nnls_n, nnls_m, nnls_a, nnls_b, w);

  /* NNLS */
  if(verbose>1) printf("applying NNLS\n");
  ret=nnls(nnls_a, nnls_m, nnls_n, nnls_b, nnls_x, &nnls_rnorm, nnls_wp, nnls_zz, nnls_index);
  if(ret>1) {
    if(verbose>1) fprintf(stderr, "NNLS solution not possible.\n");
    free(nnls_mat);
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_SOLUTION);
    return TPCERROR_NO_SOLUTION;
  }
  /* Copy a[] values */
  if(a!=NULL) for(n=0; n<nnls_n; n++) a[n]=nnls_x[n];

  /* Compute fitted y[] */
  if(yfit!=NULL) {
    if(verbose>1) printf("computing yfit[]\n");
    for(m=0; m<nnls_m; m++) {
      yfit[m]=0.0;
      for(n=0; n<nnls_n; n++)
        if(nnls_x[n]>0.0)
          yfit[m]+=nnls_x[n]*exp(-k[n]*x[m]);
    }
  }
  free(nnls_mat);

  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return TPCERROR_OK;
}
/// @cond
/*****************************************************************************/

/*****************************************************************************/
/**
 *  Main
 */
int main(int argc, char **argv)
{
  int     ai, help=0, version=0, verbose=1;
  char    tacfile[FILENAME_MAX], parfile[FILENAME_MAX], svgfile[FILENAME_MAX];
  int     weights=0; // 0=default, 1=no weighting, 2=frequency
  int     bfNr=100;
  int     expnr=0;
  int     ret;
  unsigned int model=0;


  /*
   *  Get arguments
   */
  if(argc==1) {tpcPrintUsage(argv[0], info, stderr); return(1);}
  tacfile[0]=parfile[0]=svgfile[0]=(char)0;
  /* Options */
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') {
    if(tpcProcessStdOptions(argv[ai], &help, &version, &verbose)==0) continue;
    char *cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(!*cptr) continue;
    if(strcasecmp(cptr, "W1")==0) {
      weights=1; continue;
    } else if(strcasecmp(cptr, "WF")==0) {
      weights=2; continue;
    } else if(strncasecmp(cptr, "NR=", 3)==0) {
      if(atoiCheck(cptr+3, &bfNr)==0 && bfNr>5) continue;
    } else if(strcasecmp(cptr, "5")==0 || strcasecmp(cptr, "5e")==0) {
      expnr=5; model=modelCodeIndex("5exp"); continue;
    } else if(strcasecmp(cptr, "4")==0 || strcasecmp(cptr, "4e")==0) {
      expnr=4; model=modelCodeIndex("4exp"); continue;
    } else if(strcasecmp(cptr, "3")==0 || strcasecmp(cptr, "3e")==0) {
      expnr=3; model=modelCodeIndex("3exp"); continue;
    } else if(strcasecmp(cptr, "2")==0 || strcasecmp(cptr, "2e")==0) {
      expnr=2; model=modelCodeIndex("2exp"); continue;
    } else if(strcasecmp(cptr, "1")==0 || strcasecmp(cptr, "1e")==0) {
      expnr=1; model=modelCodeIndex("1exp"); continue;
    } else if(strncasecmp(cptr, "SVG=", 4)==0 && strlen(cptr)>4) {
      strlcpy(svgfile, cptr+4, FILENAME_MAX); continue;
    }
    fprintf(stderr, "Error: invalid option '%s'.\n", argv[ai]);
    return(1);
  } else break;

  TPCSTATUS status; statusInit(&status);
  statusSet(&status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  status.verbose=verbose-3;
  
  /* Print help or version? */
  if(help==2) {tpcHtmlUsage(argv[0], info, ""); return(0);}
  if(help) {tpcPrintUsage(argv[0], info, stdout); return(0);}
  if(version) {tpcPrintBuild(argv[0], stdout); return(0);}

  /* Process other arguments, starting from the first non-option */
  if(ai<argc) strlcpy(tacfile, argv[ai++], FILENAME_MAX);
  if(ai<argc) strlcpy(parfile, argv[ai++], FILENAME_MAX);
  if(ai<argc) {
    fprintf(stderr, "Error: invalid argument '%s'.\n", argv[ai]);
    return(1);
  }
  /* Did we get all the information that we need? */
  if(!tacfile[0]) {
    fprintf(stderr, "Error: missing command-line argument; use option --help\n");
    return(1);
  }


  /* In verbose mode print arguments and options */
  if(verbose>1) {
    printf("tacfile := %s\n", tacfile);
    if(parfile[0]) printf("parfile := %s\n", parfile);
    if(svgfile[0]) printf("svgfile := %s\n", svgfile);
    if(expnr>0) printf("expnr := %d\n", expnr);
    printf("bfNr := %d\n", bfNr);
    printf("weights := %d\n", weights);
  }



  /*
   *  Read TAC data
   */
  if(verbose>1) printf("reading %s\n", tacfile);
  TAC tac; tacInit(&tac);
  ret=tacRead(&tac, tacfile, &status);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: %s\n", errorMsg(status.error));
    tacFree(&tac); return(2);
  }
  if(verbose>1) {
    printf("fileformat := %s\n", tacFormattxt(tac.format));
    printf("tacNr := %d\n", tac.tacNr);
    printf("sampleNr := %d\n", tac.sampleNr);
    printf("xunit := %s\n", unitName(tac.tunit));
    printf("yunit := %s\n", unitName(tac.cunit));
    printf("isframe := %d\n", tac.isframe);
  }
  if(tac.tacNr<1) {
    fprintf(stderr, "Error: file contains no data.\n");
    tacFree(&tac); return(2);
  }
  if(tac.tacNr>1) {
    fprintf(stderr, "Warning: file contains %d TACs; using the first one.\n", tac.tacNr);
    tac.tacNr=1;
  }
  if(tac.sampleNr<3) {
    fprintf(stderr, "Error: too few samples.\n");
    tacFree(&tac); return(2);
  }
  /* Check NaNs */
  if(tacNaNs(&tac)>0) {
    fprintf(stderr, "Error: data contains missing values.\n");
    tacFree(&tac); return(2);
  }
  if(tac.sampleNr<3) {
    fprintf(stderr, "Error: too few samples.\n");
    tacFree(&tac); return(2);
  }
  /* Sort data by sample time */
  tacSortByTime(&tac, &status);
  /* Check and set weights */
  if(weights==0) {
    if(tac.weighting==WEIGHTING_OFF || tac.weighting==WEIGHTING_UNKNOWN) {
      tac.weighting=WEIGHTING_OFF;
      for(int i=0; i<tac.sampleNr; i++) tac.w[i]=1.0;
    } 
  } else if(weights==1) {
    tac.weighting=WEIGHTING_OFF;
    for(int i=0; i<tac.sampleNr; i++) tac.w[i]=1.0;
  } else if(weights==2) {
    ret=tacWByFreq(&tac, &status);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error));
      tacFree(&tac); return(2);
    }
  }



  /*
   *  Get the first sample distance (that is > 0)
   */
  double sdist=0.0;
  {
    for(int i=1; i<tac.sampleNr; i++) {
      sdist=tac.x[i]-tac.x[0];
      if(sdist>0.0) break;
    }
    if(!(sdist>0.0)) {
      fprintf(stderr, "Error: invalid sample times.\n");
      tacFree(&tac); return(2);
    }
  }
  if(verbose>1) {
    printf("sdist := %g\n", sdist);
  }

  if(verbose>1) printf("determining initial parameter estimates\n");

  /*
   *  Spectral analysis to find better range of parameters
   */
  if(verbose>2) printf("estimating initial k value range\n");
  double kmin, kmax; // kmin must be >0
  kmin=0.000001/sdist; kmax=10.0/sdist; if(kmax>100.) kmax=100.;
  if(verbose>2) printf(" kmin := %g\n kmax := %g\n", kmin, kmax);
  double pa[bfNr], pk[bfNr], yfit[tac.sampleNr];
  ret=spectralDExp(tac.x, tac.c[0].y, tac.w, tac.sampleNr, kmin, kmax, bfNr,
                   pk, pa, yfit, &status);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: %s\n", errorMsg(status.error));
    tacFree(&tac); return(3);
  }
  if(verbose>2) {
    printf("solutions for each k:\n");
    for(int bi=0; bi<bfNr; bi++) printf("\t%g\t%g\n", pk[bi], pa[bi]);
  }
  if(verbose>5) {
    printf("measured and fitted TAC:\n");
    for(int i=0; i<tac.sampleNr; i++)
      printf("\t%g\t%g\n", tac.c[0].y[i], yfit[i]);
  }

  /* Set better k value range */
  if(verbose>2) printf("refine k value range\n");
  if(!(pa[0]>0.0 && pa[bfNr-1]>0.0)) {
    int i=0; while(i<bfNr && !(pa[i]>0.0)) kmin=pk[i++];
    i=bfNr-1; while(i>0 && !(pa[i]>0.0)) kmax=pk[i--];
    if(verbose>2) printf(" kmin := %g\n kmax := %g\n", kmin, kmax);
    ret=spectralDExp(tac.x, tac.c[0].y, tac.w, tac.sampleNr, kmin, kmax, bfNr,
                     pk, pa, yfit, &status);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error));
      tacFree(&tac); return(3);
    }
    if(verbose>2) {
      printf("solutions for each k:\n");
      for(int bi=0; bi<bfNr; bi++) printf("\t%g\t%g\n", pk[bi], pa[bi]);
    }
    if(verbose>3) {
      printf("measured and fitted TAC:\n");
      for(int i=0; i<tac.sampleNr; i++)
        printf("\t%g\t%g\n", tac.c[0].y[i], yfit[i]);
    }
  }
  if(verbose>2) {
    double v, init_wss=0.0;
    for(int i=0; i<tac.sampleNr; i++) {
      v=tac.c[0].y[i]-yfit[i];
      init_wss+=tac.w[i]*v*v;
    }
    printf(" initial_wss := %g\n", init_wss);
  }

  /* Get max a parameter to be used as basis of upper limit */
  double amax=pa[doubleMaxIndex(pa, bfNr)];
  if(verbose>3) printf("max_a := %g\n", amax);


  /* Calculate how many zero-separated basis functions there really were */
  int fbfNr=0;
  {
    int i=0, ii=0;
    do {
      i=doubleCSpanPositives(pa+ii, bfNr-ii); ii+=i; if(bfNr-ii<1) break;
      i=doubleSpanPositives(pa+ii, bfNr-ii); ii+=i; if(i>0) fbfNr++;
    } while(ii<bfNr);
  }
  if(verbose>2) printf("fbfNr := %d\n", fbfNr);
  if(fbfNr<1) {
    fprintf(stderr, "Error: cannot estimate initial parameters.\n");
    tacFree(&tac); return(3);
  }


  /* If user did not give the nr of exponentials, then set it here */
  if(expnr==0) {
    expnr=fbfNr;
    if(expnr>5) expnr=5; // there should be no need for more
    if(verbose>1) printf("expnr := %d\n", expnr);
    char s[8]; sprintf(s, "%dexp", expnr);
    model=modelCodeIndex(s);
  }


  /* Copy parameters as initial values for nonlinear optimization */
  if(verbose>2) printf("process initial parameter values\n");
  NLOPT ipar; nloptInit(&ipar);
  ret=nloptAllocate(&ipar, 2*fbfNr); // a[0], k[0], a[1], k[1], a[2], ...
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: %s\n", errorMsg(status.error));
    tacFree(&tac); return(3);
  }
  {
    int i=0, ii=0, ei=0, pi=0;
    do {
      i=doubleCSpanPositives(pa+ii, bfNr-ii); ii+=i; if(bfNr-ii<1) break;
      i=doubleSpanPositives(pa+ii, bfNr-ii);
      if(i>0) {
        if(verbose>3) {
          printf("exp function %d:\n", 1+ei);
          for(int j=0; j<i; j++) printf("  a=%g k=%g\n", pa[ii+j], pk[ii+j]);
        }
        pi=2*ei;
        ipar.xlower[pi]=ipar.xlower[pi+1]=0.0;
        ipar.xupper[pi]=10.0*amax;
        ipar.xupper[pi+1]=200.0;
        if(i==1) {
          ipar.xfull[pi]=pa[ii];   ipar.xdelta[pi]=0.1*ipar.xfull[pi];
          ipar.xfull[pi+1]=pk[ii]; ipar.xdelta[pi+1]=0.1*ipar.xfull[pi+1];
          ipar.xtol[pi]=fabs(0.005*ipar.xdelta[pi]);
          ipar.xtol[pi+1]=fabs(0.005*ipar.xdelta[pi+1]);
        } else {
          unsigned int mi=doubleMaxIndex(pa+ii, i);
          ipar.xfull[pi]=doubleSum(pa+ii, i);
          ipar.xdelta[pi]=0.1*(ipar.xfull[pi]-pa[ii+mi]);
          ipar.xfull[pi+1]=doubleMean(pk+ii, i);
          ipar.xdelta[pi+1]=0.1*(ipar.xfull[pi+1]-pk[ii+mi]);
          ipar.xtol[pi]=fabs(0.002*ipar.xdelta[pi]);
          ipar.xtol[pi+1]=fabs(0.002*ipar.xdelta[pi+1]);
        }
        ei++;
      }
      ii+=i;
    } while(ii<bfNr);
    if(verbose>5) nloptWrite(&ipar, stdout);
  }

  /* Reduce the number of exponentials, if required */
  if(ipar.totalNr>2*(unsigned int)expnr) {
    if(verbose>2)
      printf("reducing the number of exponentials from %d to %d\n", ipar.totalNr/2, expnr);
    while(ipar.totalNr>2*(unsigned int)expnr) {
      /* Search the smallest a parameter */
      int ami=0; double ama=ipar.xfull[0];
      for(unsigned int i=2; i<ipar.totalNr; i+=2)
        if(ipar.xfull[i]<ama) {ama=ipar.xfull[i]; ami=i;}
      if(verbose>5) printf("smallest a[%d]=%g\n", ami, ipar.xfull[ami]);
      /* Combine it with previous or next exponential? */
      int ci;
      if(ami==0) ci=2;
      else if(ami==(int)ipar.totalNr-2) ci=-2;
      else {if(ipar.xfull[ami-2]<ipar.xfull[ami+2]) ci=-2; else ci=2;}
      if(verbose>0) {
        printf("combining exponentials %d and %d\n", 1+ami/2, 1+(ami+ci)/2);
      }
      /* Combine a parameters */
      ipar.xfull[ami+ci]+=ipar.xfull[ami];
      ipar.xtol[ami+ci]+=ipar.xtol[ami];
      ipar.xdelta[ami+ci]*=2.0;
      /* Combine k parameters */
      ipar.xfull[ami+ci+1]=0.5*(ipar.xfull[ami+1]+ipar.xfull[ami+ci+1]);
      ipar.xtol[ami+ci+1]+=ipar.xtol[ami+1];
      ipar.xdelta[ami+ci+1]=fabs(ipar.xdelta[ami+ci+1])+fabs(ipar.xdelta[ami+1]);
      /* delete parameter from the list */
      ipar.xfull[ami]=ipar.xfull[ami+1]=nan("");
      nloptRemoveEmpties(&ipar);
      if(verbose>5) nloptWrite(&ipar, stdout);
    } 
  }

  /* Get nr of fixed parameters before deltas are modified between fittings */
  unsigned int fixedNr=nloptFixedNr(&ipar);


  /*
   *  Nonlinear optimization:
   *  Simplex
   */
  if(verbose>1) printf("nonlinear optimization\n");
  /* Set function */
  ipar._fun=func_exp;
  /* Set data pointers for the fit */
  FITDATA fitdata;
  fitdata.n=tac.sampleNr;
  fitdata.x=tac.x;
  fitdata.ymeas=tac.c[0].y;
  fitdata.ysim=yfit;
  fitdata.w=tac.w;
  ipar.fundata=&fitdata;
  if(verbose>2) printf("non-linear optimization\n");
  ret=nloptSimplex(&ipar, 0, &status);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: %s\n", errorMsg(status.error));
    tacFree(&tac); nloptFree(&ipar); return(4);
  }
  if(verbose>3) nloptWrite(&ipar, stdout);
  if(verbose>6) {
    printf("measured and fitted TAC:\n");
    for(int i=0; i<tac.sampleNr; i++)
      printf("\t%g\t%g\n", tac.c[0].y[i], yfit[i]);
  }
  if(verbose>3) {
    double v, final_wss=0.0;
    for(int i=0; i<tac.sampleNr; i++) {
      v=tac.c[0].y[i]-yfit[i];
      final_wss+=tac.w[i]*v*v;
    }
    printf(" final_wss := %g\n", final_wss);
  }
  /* Simplex again, with new deltas */
  for(unsigned int i=0; i<ipar.totalNr; i++) {
    if(fabs(ipar.xdelta[i])<1.0E-100) continue;
    if(ipar.xfull[i]>1.0E-06) ipar.xdelta[i]=0.0103*ipar.xfull[i];
    else ipar.xdelta[i]*=0.0103;
  }
  if(verbose>2) printf("non-linear optimization, 2nd time\n");
  ret=nloptSimplex(&ipar, 0, &status);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: %s\n", errorMsg(status.error));
    tacFree(&tac); nloptFree(&ipar); return(4);
  }
  if(verbose>3) nloptWrite(&ipar, stdout);
  if(verbose>6) {
    printf("measured and fitted TAC:\n");
    for(int i=0; i<tac.sampleNr; i++)
      printf("\t%g\t%g\n", tac.c[0].y[i], yfit[i]);
  }
  if(verbose>3) {
    double v, final_wss=0.0;
    for(int i=0; i<tac.sampleNr; i++) {
      v=tac.c[0].y[i]-yfit[i];
      final_wss+=tac.w[i]*v*v;
    }
    printf(" final_final_wss := %g\n", final_wss);
  }
  /* and Simplex again, with new deltas */
  for(unsigned int i=0; i<ipar.totalNr; i++) {
    ipar.xdelta[i]=0.093*ipar.xdelta[i];
  }
  if(verbose>2) printf("non-linear optimization, 3rd time\n");
  ret=nloptSimplex(&ipar, 0, &status);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: %s\n", errorMsg(status.error));
    tacFree(&tac); nloptFree(&ipar); return(4);
  }
  /* Call objective function again with the final parameters */
  (void)func_exp(ipar.totalNr, ipar.xfull, ipar.fundata);
  if(verbose>3) nloptWrite(&ipar, stdout);
  if(verbose>6) {
    printf("measured and fitted TAC:\n");
    for(int i=0; i<tac.sampleNr; i++)
      printf("\t%g\t%g\n", tac.c[0].y[i], yfit[i]);
  }
  double final_wss=0.0;
  {
    double v;
    for(int i=0; i<tac.sampleNr; i++) {
      v=tac.c[0].y[i]-yfit[i];
      final_wss+=tac.w[i]*v*v;
    }
    if(verbose>3) printf(" final_final_final_wss := %g\n", final_wss);
  }


  /*
   *  Copy function parameters into PAR struct for printing and saving
   */
  PAR par; parInit(&par);
  ret=parAllocateWithTAC(&par, &tac, ipar.totalNr, &status);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: %s\n", errorMsg(status.error));
    tacFree(&tac); nloptFree(&ipar); return(6);
  }
  par.tacNr=1; par.parNr=ipar.totalNr;
  par.format=PAR_FORMAT_TSV_UK;
  par.r[0].model=model;
  par.r[0].dataNr=tacWSampleNr(&tac);
  par.r[0].start=tac.x[0]; 
  par.r[0].end=tac.x[tac.sampleNr-1];
  par.r[0].wss=final_wss;
  par.r[0].fitNr=ipar.totalNr-fixedNr;
  /* Set parameter names and units */
  for(int i=0; i<par.parNr; i+=2) {
    sprintf(par.n[i].name, "a%d", 1+i/2);
    sprintf(par.n[i+1].name, "k%d", 1+i/2);
    par.n[i].unit=tac.cunit;
    par.n[i+1].unit=unitInverse(tac.tunit);
  }
  /* Set parameters in opposite order, and write k's as negatives */
  for(unsigned int i=0; i<ipar.totalNr; i+=2) {
    par.r[0].p[par.parNr-2-i]=ipar.xfull[i];
    par.r[0].p[par.parNr-1-i]=-ipar.xfull[i+1];
  }
  /* Print and save */
  if(verbose>0) parWrite(&par, stdout, PAR_FORMAT_TSV_UK, 1, NULL);
  if(parfile[0]) {
    /* Save file */
    if(verbose>1) printf("  saving %s\n", parfile);
    FILE *fp;
    fp=fopen(parfile, "w");
    if(fp==NULL) {
      fprintf(stderr, "Error: cannot open file for writing.\n");
      tacFree(&tac); nloptFree(&ipar); parFree(&par); return(11);
    }
    ret=parWrite(&par, fp, PAR_FORMAT_FIT, 1, &status);
    fclose(fp);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error));
      tacFree(&tac); nloptFree(&ipar); parFree(&par); return(12);
    }
    if(verbose>0) printf("parameters saved in %s\n", parfile);
  }


  /*
   *  Plot measured and fitted data, if requested
   */
  if(svgfile[0]) {
    if(verbose>1) printf("plotting measured and fitted data\n");
    /* Prepare place for function TAC */
    int snr;
    snr=(int)simSamples(0.2*sdist, 0.0, par.r[0].end-par.r[0].start, 2, NULL);
    TAC ftac; tacInit(&ftac);
    ret=tacAllocate(&ftac, snr, 1);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error));
      tacFree(&tac); nloptFree(&ipar); parFree(&par); return(21);
    }
    ftac.tacNr=1; ftac.sampleNr=snr;
    ret=tacCopyHdr(&tac, &ftac);
    if(ret==TPCERROR_OK) ret=tacCopyTacchdr(tac.c, ftac.c);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error));
      tacFree(&tac); nloptFree(&ipar); parFree(&par); tacFree(&ftac);
      return(22);
    }
    snr=(int)simSamples(0.2*sdist, 0.0, par.r[0].end-par.r[0].start, 2, ftac.x);
    ftac.isframe=0;
    for(int i=0; i<ftac.sampleNr; i++) ftac.x[i]+=par.r[0].start;
    /* Compute function values at sample times */
    ret=mfEvalY(modelCode(par.r[0].model), par.parNr, par.r[0].p, 
                ftac.sampleNr, ftac.x, ftac.c[0].y, verbose-2);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: cannot calculate function values.\n");
      tacFree(&tac); nloptFree(&ipar); parFree(&par); tacFree(&ftac);
      return(23);
    }
    if(verbose>10) tacWrite(&ftac, stdout, TAC_FORMAT_UNKNOWN, 1, NULL);
    /* Plot */
    ret=tacPlotFitSVG(&tac, &ftac, "", nan(""), nan(""), nan(""), nan(""), svgfile, &status);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error));
      tacFree(&tac); nloptFree(&ipar); parFree(&par); tacFree(&ftac);
      return(24);
    }
    tacFree(&ftac);
    if(verbose>0) printf("Measured and fitted data plotted in %s\n", svgfile);
  }


  nloptFree(&ipar);
  tacFree(&tac);
  parFree(&par);


  return(0);
}
/*****************************************************************************/

/*****************************************************************************
 *
 *  Functions to be minimized
 *
 *****************************************************************************/
double func_exp(int parNr, double *p, void *fdata)
{
  unsigned int i, ei;
  double v, wss;
  FITDATA *d=(FITDATA*)fdata;

  /* Calculate the curve values and weighted SS */
  for(i=0, wss=0.0; i<d->n; i++) {
    d->ysim[i]=0.0;
    if(isnan(d->ymeas[i]) || isnan(d->x[i])) continue;
    for(ei=0; ei<(unsigned int)parNr/2; ei++) 
      d->ysim[i]+=p[2*ei]*exp(-p[2*ei+1]*d->x[i]);
    v=d->ysim[i]-d->ymeas[i];
    wss+=d->w[i]*v*v;
  }
  if(0) {
    for(i=0; i<(unsigned int)parNr; i++) printf(" %g", p[i]);
    printf(" -> %g\n", wss);
  }

  return(wss);
}
/*****************************************************************************/

/*****************************************************************************/
/// @endcond
