/** @file avgttac.c
 *  @brief Calculate average(s) from TTACs in separate files with equal sample
 *  times.
 *  @copyright (c) Turku PET Centre
 *  @author Vesa Oikonen
 */
/// @cond
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
#include "tpcextensions.h"
#include "tpcift.h"
#include "tpccsv.h"
#include "tpctac.h"
#include "tpcstatist.h"
/*****************************************************************************/

/*****************************************************************************/
static char *info[] = {
  "Calculate average TTAC(s) from several PET TTAC files with equal sample",
  "(frame) times and regions. Each region is processed separately.",
  "TTACs are not scaled in this process, therefore user must scale the data",
  "to for example SUV or %ID before using this program.",
  " ",
  "Usage: @P [Options] meanfile tacfiles",
  " ",
  "Options:",
  " -sd=<filename>",
  "     Regional standard deviations are written into given datafile.",
  " --force",
  "     Program does not mind if the time or calibration units do not match",
  "     or files contain variable number of TACs."
  " -stdoptions", // List standard options like --help, -v, etc
  " ",
  "If only one input datafile is given, it is assumed to contain a list of",
  "bolus datafiles with paths if necessary. Tabs, commas and newlines can be",
  "used to separate filenames in the list file.",
  " ",
  "Example 1:",
  "    avgttac meansuv.dft *suv.dft",
  "Example 2:",
  "    dir /b *suv.dft > filelist.txt",
  "    avgttac meansuv.dft filelist.txt",
  " ",
  "See also: avgbolus, avgfract, dftavg, dftsuv, tacsort, taccut, tacren",
  " ",
  "Keywords: TAC, average, modelling, simulation",
  0};
/*****************************************************************************/

/*****************************************************************************/
/* Turn on the globbing of the command line, since it is disabled by default in
   mingw-w64 (_dowildcard=0); in MinGW32 define _CRT_glob instead, if necessary;
   In Unix&Linux wildcard command line processing is enabled by default. */
/*
#undef _CRT_glob
#define _CRT_glob -1
*/
int _dowildcard = -1;
/*****************************************************************************/

/*****************************************************************************/
/**
 *  Main
 */
int main(int argc, char **argv)
{
  int     ai, help=0, version=0, verbose=1;
  char   *cptr, outfile[FILENAME_MAX], sdfile[FILENAME_MAX];
  int     forceMode=0;
  int     ret, fileNr;
  CSV     csv;

  
  /*
   *  Get arguments
   */
  if(argc==1) {tpcPrintUsage(argv[0], info, stderr); return(1);}
  outfile[0]=sdfile[0]=(char)0;
  csvInit(&csv);
  /* Options */
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') {
    if(tpcProcessStdOptions(argv[ai], &help, &version, &verbose)==0) continue;
    cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(!*cptr) continue;
    if(strncasecmp(cptr, "SD=", 3)==0) {
      if(strlcpy(sdfile, cptr+3, FILENAME_MAX)>0) continue;
    } else if(strncasecmp(cptr, "FORCE", 1)==0) {
      forceMode=1; continue;
    }
    fprintf(stderr, "Error: invalid option '%s'.\n", argv[ai]);
    return(1);
  } else break; // tac name argument may start with '-'

  TPCSTATUS status; statusInit(&status);
  statusSet(&status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  status.verbose=verbose-4;
  
  /* Print help or version? */
  if(help==2) {tpcHtmlUsage(argv[0], info, ""); return(0);}
  if(help) {tpcPrintUsage(argv[0], info, stdout); return(0);}
  if(version) {tpcPrintBuild(argv[0], stdout); return(0);}

  /* Process other arguments, starting from the first non-option */
  /* First one is the filename for averages */ 
  if(ai<argc) {strlcpy(outfile, argv[ai], FILENAME_MAX); ai++;}
  if(ai==argc) {
    fprintf(stderr, "Error: missing command-line argument; use option --help\n");
    return(1);
  }
  /* If only one argument left, then that is the filename of file list */
  if(ai==argc-1) {
    if(verbose>1) printf("reading filename list in %s\n", argv[ai]);
    FILE *fp;
    fp=fopen(argv[ai], "r"); if(fp==NULL) {
      fprintf(stderr, "Error: cannot open file %s\n", argv[ai]); return(1);}
    ret=csvRead(&csv, fp, &status); fclose(fp);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error)); return(1);}
  } else {
    if(verbose>1) printf("filenames listed as arguments\n");
    for(; ai<argc; ai++) csvPutString(&csv, argv[ai], 1);
  }

  /* In verbose mode print arguments and options */
  if(verbose>1) {
    printf("outfile := %s\n", outfile);
    if(sdfile[0]) printf("sdfile := %s\n", sdfile);
    printf("forceMode := %d\n", forceMode);
    printf("files:\n"); csvWrite(&csv, stdout, NULL);
  }

  /* Check that files exist, and that there are at least two of them */
  fileNr=0;
  for(int i=0; i<csv.nr; i++) {
    if(access(csv.c[i].content, 0) == -1) {
      fprintf(stderr, "Error: file '%s' is not accessible.\n", csv.c[i].content);
      csvFree(&csv); return(2);
    }
    fileNr++;
  }
  if(verbose>1) printf("fileNr := %d\n", fileNr);
  if(fileNr<2) {
    fprintf(stderr, "Error: too few TAC files for averaging.\n");
    csvFree(&csv); return(2);
  }



  /*
   *  Read all TTAC files
   */
  /* Allocate memory for an array of TACs */
  TAC *taclist;
  taclist=(TAC*)malloc(fileNr*sizeof(TAC));
  if(taclist==NULL) {
    fprintf(stderr, "Error: out of memory.\n");
    csvFree(&csv); return(3);
  }
  /* Initiate TAC structs */
  for(int i=0; i<fileNr; i++) tacInit(taclist+i);
  /* Read TTAC files */
  for(int i=0; i<fileNr; i++) {
    if(verbose>1) printf("reading %s\n", csv.c[i].content);
    if(tacRead(taclist+i, csv.c[i].content, &status)) {
      fprintf(stderr, "Error: %s\n", errorMsg(status.error));
      for(int j=0; j<i; j++) tacFree(taclist+j);
      csvFree(&csv); free(taclist); return(3);
    }
    if(verbose>2) {
      printf("fileformat := %s\n", tacFormattxt(taclist[i].format));
      printf("tacNr := %d\n", taclist[i].tacNr);
      printf("sampleNr := %d\n", taclist[i].sampleNr);
    }
    /* Check that data does not have missing values that cannot be fixed */
    if(tacFixNaNs(taclist+i)) {
      fprintf(stderr, "Warning: missing sample(s) in %s\n", csv.c[i].content);
    }
  }

  if(verbose>1) printf("checking data...\n");

  /* Check that TAC files have at least as many regions and samples as 
     the first file has */
  int tacNr, sampleNr;
  {
    int n=0, m=0;
    tacNr=taclist[0].tacNr; sampleNr=taclist[0].sampleNr;
    for(int i=1; i<fileNr; i++) {
      if(taclist[i].tacNr!=tacNr) {
        m++; if(taclist[i].tacNr<tacNr) tacNr=taclist[i].tacNr;
      }
      if(taclist[i].sampleNr!=sampleNr) {
        n++; if(taclist[i].sampleNr<sampleNr) sampleNr=taclist[i].sampleNr;
      }
    }
    if(!forceMode && (n>0 || m>0)) {
      if(n>0) fprintf(stderr, "Error: different nr of samples.\n");
      if(m>0) fprintf(stderr, "Error: different nr of regions.\n");
      for(int i=0; i<fileNr; i++) tacFree(taclist+i);
      csvFree(&csv); free(taclist); return(3);
    }
    if(n>0) fprintf(stderr, "Warning: different nr of samples.\n");
    if(m>0) fprintf(stderr, "Warning: different nr of regions.\n");
  }
  if(verbose>1) {
    printf("commonSampleNr := %d\n", sampleNr);
    printf("commonTacNr := %d\n", tacNr);
  }
  
  /* Convert TAC units when necessary */
  for(int i=1; i<fileNr; i++) {
    ret=tacXUnitConvert(taclist+i, taclist[0].tunit, &status);
    if(ret==TPCERROR_UNKNOWN_UNIT) ret=TPCERROR_OK;
    if(ret==TPCERROR_OK) {
      ret=tacYUnitConvert(taclist+i, taclist[0].cunit, &status);
      if(ret==TPCERROR_UNKNOWN_UNIT) ret=TPCERROR_OK;
    }
    if(ret!=TPCERROR_OK) { // failed
      /* failing is a problem if units are to be verified */
      if(!forceMode) {
        fprintf(stderr, "Error: TAC units do match.\n");
        for(int i=0; i<fileNr; i++) tacFree(taclist+i);
        csvFree(&csv); free(taclist); return(4);
      }
    }    
  }

  /* Check region names, if more than one TAC per file */
  if(!forceMode && tacNr>1) {
    for(int i=1; i<fileNr; i++) {
      for(int j=0; j<tacNr; j++) {
        if(tacCompareNames(taclist, taclist+i, j, &status)) {
          fprintf(stderr, "Error: TAC names do match.\n");
          for(int i=0; i<fileNr; i++) tacFree(taclist+i);
          csvFree(&csv); free(taclist); return(5);
        }
      }
    }
  }



  /*
   *  Calculate mean TACs
   */
  if(verbose>1) printf("calculating mean...\n");
  /* Allocate memory for mean TACs */
  TAC tacmean, tacsd;
  tacInit(&tacmean); tacInit(&tacsd); 
  if(tacAllocate(&tacmean, sampleNr, tacNr)!=TPCERROR_OK) {
    fprintf(stderr, "Error: cannot allocate memory for mean TACs.\n");
    for(int i=0; i<fileNr; i++) tacFree(taclist+i);
    csvFree(&csv); free(taclist); return(6);
  }
  tacmean.sampleNr=sampleNr; tacmean.tacNr=tacNr;
  ret=tacCopyHdr(taclist, &tacmean);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: cannot copy TAC header.\n");
    for(int i=0; i<fileNr; i++) tacFree(taclist+i);
    csvFree(&csv); free(taclist); tacFree(&tacmean); return(6);
  }
  tacmean.weighting=WEIGHTING_OFF;
  tacSetHeaderStudynr(&tacmean.h, "Mean");
  for(int i=0; i<tacNr; i++) tacCopyTacchdr(taclist[0].c+i, tacmean.c+i);
  /* Copy sample times */
  ret=tacXCopy(taclist, &tacmean, 0, sampleNr-1);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error: cannot copy sample times.\n");
    for(int i=0; i<fileNr; i++) tacFree(taclist+i);
    csvFree(&csv); free(taclist); tacFree(&tacmean); return(6);
  }
  /* Allocate place for SDs too, if required */
  if(sdfile[0]) {
    if(tacDuplicate(&tacmean, &tacsd)!=TPCERROR_OK) {
      fprintf(stderr, "Error: cannot allocate memory for SD data.\n");
      for(int i=0; i<fileNr; i++) tacFree(taclist+i);
      csvFree(&csv); free(taclist); tacFree(&tacmean); return(6);
    }
    tacSetHeaderStudynr(&tacsd.h, "SD");
  }
  /* Compute the mean, and SD if required */
  for(int i=0; i<tacNr; i++) {
    double tempd[fileNr], mean, sd;
    for(int j=0; j<sampleNr; j++) {
      for(int k=0; k<fileNr; k++) tempd[k]=taclist[k].c[i].y[j];
      statMeanSD(tempd, fileNr, &mean, &sd, NULL);
      tacmean.c[i].y[j]=mean;
      if(sdfile[0]) tacsd.c[i].y[j]=sd;
    }
    /* also ROI size */
    for(int k=0; k<fileNr; k++) tempd[k]=taclist[k].c[i].size;
    statMeanSD(tempd, fileNr, &mean, &sd, NULL);
    tacmean.c[i].size=mean;
    if(sdfile[0]) tacsd.c[i].size=sd;
  }

  for(int i=0; i<fileNr; i++) tacFree(taclist+i);
  free(taclist);
  csvFree(&csv);


  /* 
   *  Save data 
   */
  if(verbose>1) printf("writing %s\n", outfile);
  FILE *fp; fp=fopen(outfile, "w");
  if(fp==NULL) {
    fprintf(stderr, "Error: cannot open file for writing.\n");
    tacFree(&tacmean); tacFree(&tacsd);
    return(11);
  }
  ret=tacWrite(&tacmean, fp, TAC_FORMAT_UNKNOWN, 1, &status);
  fclose(fp);
  if(ret!=TPCERROR_OK) {
    fprintf(stderr, "Error (%d): %s\n", ret, errorMsg(status.error));
    tacFree(&tacmean); tacFree(&tacsd); return(12);
  }
  if(verbose>0) printf("Mean TAC(s) written in %s\n", outfile);
  if(sdfile[0]) {
    if(verbose>1) printf("writing %s\n", sdfile);
    FILE *fp; fp=fopen(sdfile, "w");
    if(fp==NULL) {
      fprintf(stderr, "Error: cannot open file for writing.\n");
      tacFree(&tacmean); tacFree(&tacsd);
      return(21);
    }
    ret=tacWrite(&tacsd, fp, TAC_FORMAT_UNKNOWN, 1, &status);
    fclose(fp);
    if(ret!=TPCERROR_OK) {
      fprintf(stderr, "Error (%d): %s\n", ret, errorMsg(status.error));
      tacFree(&tacmean); tacFree(&tacsd); return(22);
    }
    if(verbose>0) printf("TAC SDs written in %s\n", sdfile);
  }

  /* Free memory */
  tacFree(&tacmean); tacFree(&tacsd);

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/// @endcond

#if(0)
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
#include "curveio.h"
#include "libtpcmodext.h"
#include "libtpcmisc.h"
#include "libtpcmodel.h"
/*****************************************************************************/
#include "readfile.h"
/*****************************************************************************/
int main(int argc, char *argv[])
{
  int ai, bi, fi, ri, rj, ret, n, m, dftNr=0, voiNr, frameNr;
  int save_errors=0;
  char *cptr, outfile[FILENAME_MAX], tmp[FILENAME_MAX];
  STR_TOKEN_LIST filelist;
  DFT *dftlist, dft;
  double sd;


  /*
   *  Get arguments
   */
  if(argc==1) {print_usage(); return(1);}
  str_token_list_init(&filelist);
  outfile[0]=(char)0;
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') { /* Options */
    cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(cptr==NULL) continue;
    if(strncasecmp(cptr, "TEST", 4)==0) {
      cptr+=4; if(*cptr=='=') TEST=atoi(cptr+1); else TEST=1; continue;
    } else if(strcasecmp(cptr, "V")==0 || strcasecmp(cptr, "VERSION")==0 ||
              strcasecmp(cptr, "BUILD")==0) {
      print_build();
      return(0);
    } else if(strcasecmp(cptr, "H")==0 || strcasecmp(cptr, "HELP")==0) {
      print_usage(); return(0);
    } else if(strcasecmp(cptr, "SILENT")==0) {
      SILENT=1; VERBOSE=0; continue;
    } else if(strcasecmp(cptr, "VERBOSE")==0) {
      VERBOSE=1; SILENT=0; continue;
    } else if(strcasecmp(cptr, "SD")==0) {
      save_errors=1; continue;
    }
    fprintf(stderr, "Error: invalid option '%s'.\n", argv[ai]);
    return(1);
  } else { /* Not an option, but filename */
    if(!outfile[0]) {
      strcpy(outfile, argv[ai]); continue;
    } else {
      if(str_token_list_add(&filelist, argv[ai])) {
        fprintf(stderr, "Error: invalid argument '%s'.\n", argv[ai]);
        return(1);
      }
    }
  }
  if(!outfile[0] || filelist.token_nr<1) {
    fprintf(stderr, "Error: missing command-line argument; try %s --help\n",
      argv[0]);
    return(1);
  }
  /* Print arguments for testing */
  if(TEST) {
    printf("PROGRAM := %s\n", PROGRAM);
    printf("TEST := %d\n", TEST);
    printf("SILENT := %d\n", SILENT);
    printf("VERBOSE := %d\n", VERBOSE);
    printf("save_errors := %d\n", save_errors);
    printf("token_nr := %d\n", filelist.token_nr);
    printf("outfile := %s\n", outfile);
    fflush(stdout);
  }
  if(TEST>0 && VERBOSE==0) VERBOSE=1;

  /*
   *  Check filename list; read list of files if necessary
   */
  if(filelist.token_nr>1) {
    /* Filenames were given directly */
    if(TEST) printf("filenames were given on command-line\n");
  } else { /* Filenames were given in a list file */
    strcpy(tmp, filelist.tok[0]);
    if(access(tmp, 0) == -1) {
      fprintf(stderr, "Error: '%s' is not accessible.\n", tmp); return(2);}
    str_token_list_empty(&filelist);
    if(VERBOSE) printf("reading filenames in %s\n", tmp);
    ret=str_token_list_read(tmp, &filelist);
    if(ret) {
      fprintf(stderr, "Error %d in reading filename list %s\n", ret, tmp);
      return(2);
    }
  }
  dftNr=filelist.token_nr;
  if(TEST) printf("dftNr := %d\n", dftNr);
  if(dftNr<2) {
    fprintf(stderr, "Error: missing command-line argument; try %s --help\n",
      argv[0]);
    str_token_list_empty(&filelist); return(1);
  }
  if(TEST) { /* List filenames */
    for(bi=0; bi<dftNr; bi++)
      fprintf(stdout, "datafile[%d] := %s\n", bi+1, filelist.tok[bi]);
  }

  /* Check that files exist */
  for(bi=0; bi<dftNr; bi++) {
    if(access(filelist.tok[bi], 0) == -1) {
      fprintf(stderr, "Error: file '%s' is not accessible.\n",
              filelist.tok[bi]);
      str_token_list_empty(&filelist); return(3);
    }
  }


  /*
   *  Read datafiles
   */
  /* Allocate memory for an array of DFT data */
  dftlist=(DFT*)malloc(dftNr*sizeof(DFT));
  if(dftlist==NULL) {
    fprintf(stderr, "Error: out of memory.\n");
    str_token_list_empty(&filelist); return(4);
  }
  /* Read bolus datafiles */
  for(bi=0; bi<dftNr; bi++) {
    if(VERBOSE) printf("reading %s\n", filelist.tok[bi]);
    dftInit(dftlist+bi);
    if(dftRead(filelist.tok[bi], dftlist+bi)) {
      fprintf(stderr, "Error in reading '%s': %s\n", filelist.tok[bi], dfterrmsg);
      str_token_list_empty(&filelist); free(dftlist); return(5);
    }
    if(TEST) printf(" -> %d frames and %d curves\n",
      dftlist[bi].frameNr, dftlist[bi].voiNr);
  }
  /* Get the smallest frame and ROI nr */
  voiNr=dftlist[0].voiNr; frameNr=dftlist[0].frameNr; n=m=0;
  for(bi=1; bi<dftNr; bi++) {
    if(dftlist[bi].voiNr<voiNr) {voiNr=dftlist[bi].voiNr; m++;}
    if(dftlist[bi].frameNr<frameNr) {frameNr=dftlist[bi].frameNr; n++;}
  }
  if(n>0) fprintf(stderr, "Warning: different nr of samples.\n");
  if(m>0) fprintf(stderr, "Warning: different nr of regions.\n");
  if(TEST) {
    printf("frameNr := %d\n", frameNr);
    printf("voiNr := %d\n", voiNr);
  }
  if(VERBOSE) printf("checking data...\n");
  /* Check that concentrations units are the same in all TAC files */
  // NOT APPLIED in this release.
  /* Check that sample times are equal in all TAC files */
  for(bi=1; bi<dftNr; bi++) {
    /* This routine works if time units or frame nr are different */
    if(check_times_dft_vs_dft(dftlist, dftlist+bi, TEST-2)!=1) {
      fprintf(stderr, "Error: different sample times in %s\n", filelist.tok[bi]);
      str_token_list_empty(&filelist);
      for(bi=dftNr-1; bi>=0; bi--) dftEmpty(dftlist+bi);
      free(dftlist); return(6);
    }
  }
  /* Check that ROI names are the same in all TAC files */
  if(voiNr>1) { /* but only when files contain more than one TAC */
    for(bi=1; bi<dftNr; bi++) {
      for(ri=0; ri<voiNr; ri++) {
        if(strcasecmp(dftlist[0].voi[ri].name, dftlist[bi].voi[ri].name)==0)
          continue;
        if(VERBOSE) {
          printf("dft[%d].voi[%d].name := '%s'\n",
                  0, ri, dftlist[0].voi[ri].name);
          printf("dft[%d].voi[%d].name := '%s'\n",
                  bi, ri, dftlist[bi].voi[ri].name);
        }
        fprintf(stderr, "Error: different region names in %s\n",
                filelist.tok[bi]);
        str_token_list_empty(&filelist);
        for(bi=dftNr-1; bi>=0; bi--) dftEmpty(dftlist+bi);
        free(dftlist);
        return(6);
      }
    }
  }  
  /* Check for NaN's */
  for(bi=0; bi<dftNr; bi++) {
    ret=dft_nr_of_NA(dftlist+bi);
    if(ret>0)
      fprintf(stderr, "Warning: missing sample(s) in %s\n", filelist.tok[bi]);
  }

  /*
   *  Calculate mean TACs
   */
  if(VERBOSE) printf("calculating mean...\n");
  /* Allocate memory for mean TACs */
  dftInit(&dft); m=voiNr; if(save_errors) m*=2;
  ret=dftSetmem(&dft, frameNr, m);
  if(ret) {
    fprintf(stderr, "Error (%d): cannot allocate memory for mean TACs.\n", ret);
    str_token_list_empty(&filelist);
    for(bi=dftNr-1; bi>=0; bi--) dftEmpty(dftlist+bi);
    free(dftlist);
    return(7);
  }
  dft.frameNr=frameNr; dft.voiNr=m;
  /* Copy times & header info */
  dftCopymainhdr(dftlist, &dft);
  for(fi=0; fi<frameNr; fi++) {
    dft.x[fi]=dftlist[0].x[fi];
    dft.x1[fi]=dftlist[0].x1[fi]; dft.x2[fi]=dftlist[0].x2[fi];
  }
  for(ri=0; ri<voiNr; ri++) {
    if(!save_errors) {
      dftCopyvoihdr(dftlist, ri, &dft, ri);
    } else {
      dftCopyvoihdr(dftlist, ri, &dft, 2*ri);
      dftCopyvoihdr(dftlist, ri, &dft, 2*ri+1);
      strcpy(dft.voi[2*ri+1].name, "SD");
      strcpy(dft.voi[2*ri+1].voiname, "SD");
      strcpy(dft.voi[2*ri+1].hemisphere, "");
      strcpy(dft.voi[2*ri+1].place, "");
    }
  }
  strcpy(dft.studynr, "mean");
  if(TEST>10) dftPrint(&dft);
  /* Compute the mean, and SD if required */
  double tempd[dftNr];
  for(ri=0; ri<voiNr; ri++) {
    rj=ri; if(save_errors) rj=2*ri;
    for(fi=0; fi<dft.frameNr; fi++) {
      for(bi=0; bi<dftNr; bi++) tempd[bi]=dftlist[bi].voi[ri].y[fi];
      dft.voi[rj].y[fi]=dmean_nan(tempd, dftNr, &sd, &n);
      if(save_errors) dft.voi[rj+1].y[fi]=sd;
    } // next sample
    /* Average of ROI volumes */
    for(bi=0; bi<dftNr; bi++) tempd[bi]=dftlist[bi].voi[ri].size;
    dft.voi[rj].size=dmean_nan(tempd, dftNr, &sd, &n);
    if(save_errors) dft.voi[rj+1].size=sd;
  } // next region

  /*
   *  Save mean data
   */
  if(VERBOSE) printf("saving mean TACs in %s\n", outfile);
  /* Write DFT file */
  if(dftWrite(&dft, outfile)) {
    fprintf(stderr, "Error in writing '%s': %s\n", outfile, dfterrmsg);
    str_token_list_empty(&filelist);
    for(bi=dftNr-1; bi>=0; bi--) dftEmpty(dftlist+bi);
    free(dftlist); dftEmpty(&dft);
    return(11);
  }
  if(SILENT==0) fprintf(stdout, "Mean TAC(s) written in %s\n", outfile);


  /* Free memory */
  str_token_list_empty(&filelist);
  for(bi=dftNr-1; bi>=0; bi--) dftEmpty(dftlist+bi);
  free(dftlist); dftEmpty(&dft);

  return(0);
}
#endif
